import json
from collections import namedtuple

from scipy.sparse import csr_matrix
from sklearn.linear_model import LogisticRegression
from sklearn.feature_extraction.text import CountVectorizer
import treform as ptm

KeywordScore = namedtuple('KeywordScore', 'word frequency coefficient')

class LassoTermExtractor:

    def __init__(self, min_tf=20, min_df=10, costs=None, verbose=True):
        name = 'lasso'
        self.min_tf = min_tf
        self.min_df = min_df
        self.costs = [500, 200, 100, 50, 10, 5, 1, 0.1] if costs == None else costs
        self.costs = sorted(self.costs)
        self.verbose = verbose
        self.index2word = None
        self.word2index = None
        self._tfs = {}

    def train(self, x, index2word=None, word2index=None):
        self.num_doc, self.num_term = x.shape
        self.index2word = index2word
        self.word2index = word2index

        rows, cols = x.nonzero()
        b = csr_matrix(([1] * len(rows), (rows, cols)))
        _df = dict(enumerate(b.sum(axis=0).tolist()[0]))
        _df = {word: df for word, df in _df.items() if df >= self.min_df}

        self._tfs = dict(enumerate(x.sum(axis=0).tolist()[0]))
        self._tfs = {word: freq for word, freq in self._tfs.items() if (freq >= self.min_df) and (word in _df)}

        rows_ = []
        cols_ = []
        data_ = []
        for r, c, d in zip(rows, cols, x.data):
            if not (c in self._tfs):
                continue
            rows_.append(r)
            cols_.append(c)
            data_.append(d)
        self.x = csr_matrix((data_, (rows_, cols_)))
        self._is_empty = [1 if float(d[0]) == 0 else 0 for d in self.x.sum(axis=1)]

    def extract_from_word(self, word, min_num_of_keywords=5):
        word = self._encoding(word)
        pos_idx = self.get_document_index(word)
        return self.extract_from_docs(pos_idx, min_num_of_keywords, except_words={word})

    def _encoding(self, word):
        if type(word) == str:
            if not self.word2index:
                raise ValueError('You should set index2word first')
            word = self.word2index.get(word, -1)
        return word

    def get_document_index(self, word):
        word = self._encoding(word)
        if not (0 <= word < self.num_term):
            return []
        return self.x[:, word].nonzero()[0].tolist()

    def extract_from_docs(self, docs_idx, min_num_of_keywords=5, except_words=None):
        pos_idx = set(docs_idx)
        print(str(len(pos_idx)))
        #for d in pos_idx:
            #print(str(self._is_empty[d]) + " :: " + str(d))
        y = [1 if (d in pos_idx and self._is_empty[d] == 0) else -1 for d in range(self.num_doc)]
        if (1 in set(y)) == False:
            if self.verbose:
                print('There is no corresponding documents')
            return []

        if except_words:
            data_ = []
            rows_ = []
            cols_ = []
            rows, cols = self.x.nonzero()
            for r, w, d in zip(rows, cols, self.x.data):
                if w in except_words:
                    continue
                data_.append(d)
                rows_.append(r)
                cols_.append(w)
            x_ = csr_matrix((data_, (rows_, cols_)))
        else:
            x_ = self.x

        for c in self.costs:
            logistic = LogisticRegression(penalty='l1', C=c, solver='liblinear')
            logistic.fit(x_, y)
            coefficients = logistic.coef_.reshape(-1)
            keywords = sorted(enumerate(coefficients), key=lambda x: x[1], reverse=True)
            keywords = [(word, self._tfs.get(word, 0), coef) for word, coef in keywords if coef > 0]
            logistic = None
            if self.verbose:
                print('%d keywords extracted from %.3f cost' % (len(keywords), c))
            if len(keywords) >= min_num_of_keywords:
                break
        if self.index2word:
            keywords = [KeywordScore(self.index2word[word] if 0 <= word < self.num_term else 'Unk%d' % word, tf, coef)
                        for word, tf, coef in keywords]
        else:
            keywords = [KeywordScore(word, tf, coef) for word, tf, coef in keywords]
        return keywords


if __name__ == '__main__':

    input_file = './data/3_class_naver_news.csv'
    # 1. text processing and representation
    corpus = ptm.CorpusFromFieldDelimitedFileForClassification(input_file,
                                                               delimiter=',',
                                                               doc_index=4,
                                                               class_index=1,
                                                               title_index=3)
    tups = corpus.pair_map
    class_list = []
    for id in tups:
        # print(tups[id])
        class_list.append(tups[id])

    mecab_path = 'C:\\mecab\\mecab-ko-dic'
    pipeline = ptm.Pipeline(ptm.splitter.NLTK(),
                            ptm.tokenizer.MeCab(mecab_path),
                            ptm.helper.POSFilter('NN*'),
                            ptm.helper.SelectWordOnly(),
                            ptm.ngram.NGramTokenizer(1, 2),
                            ptm.helper.StopwordFilter(file='../stopwords/stopwordsKor.txt')
                            )
    result = pipeline.processCorpus(corpus)
    print('==  ==')

    with open("./model/id_to_category.json") as handle:
        id_to_category = json.loads(handle.read())

    # 0 - economy 1 - IT 2 - politics
    category = []
    documents = []
    idx = 0
    for doc in result:
        document = ''
        for sent in doc:
            document += " ".join(sent)
        documents.append(document)
        label = class_list[idx]

        if label == id_to_category[str(2)]:
            category.append(idx)
        idx += 1

    vectorizer = CountVectorizer()
    x = vectorizer.fit_transform(documents)
    print(x.shape)

    word2index = vectorizer.vocabulary_
    index2word = sorted(
        vectorizer.vocabulary_,
        key=lambda x: vectorizer.vocabulary_[x]
    )

    lasso = LassoTermExtractor(costs=[500, 200, 100, 50, 10, 5, 1, 0.1],
                               min_tf=10,
                               min_df=5)
    lasso.train(x,index2word,word2index)
    #keywords = lasso.extract_from_word(137,min_num_of_keywords=30)

    keywords = lasso.extract_from_docs(category,min_num_of_keywords=30)
    print(keywords[:20])



#!/bin/env python
#-*- coding: utf8 -*-

import tensorflow as tf

def weight_variable(shape):
    initial = tf.truncated_normal(shape, stddev=0.1)
    return tf.Variable(initial)

def bias_variable(shape):
    initial = tf.constant(0.1, shape=shape)
    return tf.Variable(initial)

def RNN(_X, _istate, _weights, _biases, n_hidden, n_steps, n_input, early_stop):
    # input _X shape: Tensor("Placeholder:0", shape=(?, n_steps, n_input), dtype=float32)
    # switch n_steps and batch_size, Tensor("transpose:0", shape=(n_steps, ?, n_input), dtype=float32)
    _X = tf.transpose(_X, [1, 0, 2])
    # Reshape to prepare input to hidden activation
    # (n_steps*batch_size, n_input) => (?, n_input), Tensor("Reshape:0", shape=(?, n_input), dtype=float32)
    _X = tf.reshape(_X, [-1, n_input])
    # Linear activation
    _X = tf.matmul(_X, _weights['hidden']) + _biases['hidden'] # (?, n_hidden)+scalar(n_hidden,)=(?,n_hidden)

    # Define a lstm cell with tensorflow
    lstm_cell = tf.contrib.rnn.LSTMCell(n_hidden, forget_bias=1.0, state_is_tuple=False)
    # Split data because rnn cell needs a list of inputs for the RNN inner loop
    # n_steps splits each of which contains (?, n_hidden)
    # ex) [<tf.Tensor 'split:0' shape=(?, n_hidden) dtype=float32>, ... , <tf.Tensor 'split:n_steps-1' shape=(?, n_hidden) dtype=float32>]
    _X = tf.split(_X, n_steps, 0)
    # Get lstm cell output
    outputs, states = tf.contrib.rnn.static_rnn(cell=lstm_cell, inputs=_X, initial_state=_istate, sequence_length=early_stop)
    final_outputs = []
    for output in outputs :
        # Linear activation
        final_output = tf.matmul(output, _weights['out']) + _biases['out'] # (?, n_classes)
        final_outputs.append(final_output)
    # [<tf.Tensor 'add_1:0' shape=(?, n_classes), ..., <tf.Tensor 'add_n_steps:0' shape=(?, n_classes) dtype=float32>]
    return final_outputs


from treform.spelling import *
import treform
from treform.spelling.config import Config

import os

from treform.utility.dataset import DataSet
from treform.utility.num_util import NumUtil
from treform.utility.char_one_hot_vector import CharOneHotVector
from treform.utility.datafile_util import DataFileUtil
from treform.spelling.spelling_error_correction_upgraded import SpellingErrorCorrection
import tensorflow as tf


class BaseSpellingCorrector:
    IN_TYPE = [str]
    OUT_TYPE = [str]

class DAESpellingCorrector(BaseSpellingCorrector):
    IN_TYPE = [str]
    OUT_TYPE = [str]

    def __init__(self, json_file=''):
        config = Config(json_file)
        KO_WIKIPEDIA_ORG_SPELLING_ERROR_CORRECTION_MODEL_DIR = config.KO_WIKIPEDIA_ORG_SPELLING_ERROR_CORRECTION_MODEL_DIR
        KO_WIKIPEDIA_ORG_TRAIN_SENTENCES_FILE = config.KO_WIKIPEDIA_ORG_TRAIN_SENTENCES_FILE
        KO_WIKIPEDIA_ORG_VALID_SENTENCES_FILE = config.KO_WIKIPEDIA_ORG_VALID_SENTENCES_FILE
        KO_WIKIPEDIA_ORG_TEST_SENTENCES_FILE = config.KO_WIKIPEDIA_ORG_TEST_SENTENCES_FILE
        KO_WIKIPEDIA_ORG_CHARACTERS_FILE = config.KO_WIKIPEDIA_ORG_CHARACTERS_FILE
        KO_WIKIPEDIA_ORG_DIR = config.KO_WIKIPEDIA_ORG_DIR
        self.window_size = 10
        characters_file = KO_WIKIPEDIA_ORG_CHARACTERS_FILE
        #D:\python_workspace\treform\treform\spelling\models\spelling_error_correction\spelling_error_correction_model.sentences=3.window_size=10.noise_rate=0.1.n_hidden=100
        #self.model_file = os.path.join(KO_WIKIPEDIA_ORG_SPELLING_ERROR_CORRECTION_MODEL_DIR,
        #                            'spelling_error_correction_model.sentences=3.window_size=10.noise_rate=0.1.n_hidden=100')

        self.A = config.SPELLING_MODEL_DIR
        self.model_file = config.SPELLING_MODEL_DIR + "/model"
        self.model_meta_file = config.SPELLING_MODEL_DIR + "/model.meta"
        self.features_vector = CharOneHotVector(DataFileUtil.read_list(characters_file))
        labels_vector = CharOneHotVector(DataFileUtil.read_list(characters_file))
        self.n_features = len(self.features_vector) * self.window_size  # number of features
        n_classes = len(labels_vector) * self.window_size
        self.total_epoch = 5
        self.features_vector_size = self.n_features // self.window_size

    def __call__(self, *args, **kwargs):
        #model_path = "/home/bible/treform/spelling_training/models/spelling_error_correction/spelling_error_correction_model.sentences=3.window_size=10.noise_rate=0.1.n_hidden=100/model"

        tf.compat.v1.disable_v2_behavior()
        tf.compat.v1.disable_eager_execution()

        detection_graph = tf.compat.v1.get_default_graph()

        with detection_graph.as_default() as default_graph:

            with tf.compat.v1.Session() as sess:
                # Load the graph with the trained states
                loader = tf.compat.v1.train.import_meta_graph(self.model_meta_file)
                loader.restore(sess, self.model_file)

                # Get the tensors by their variable name
                # Make predictions
                dropout_keep_rate = 1.0
                dropout_keep_prob = tf.compat.v1.placeholder(tf.compat.v1.float32)
                X = detection_graph.get_tensor_by_name('X:0')  # shape=(batch_size, window_size * feature_vector.size)
                # X = tf.cast(X, dtype='float32')
                print(X)
                Y = detection_graph.get_tensor_by_name('Y:0')
                W1 = detection_graph.get_tensor_by_name('W1:0')
                print(W1)
                # W1 = tf.cast(W1, dtype='float32')
                b1 = detection_graph.get_tensor_by_name('b1:0')
                print(b1)

                # op = sess.graph.get_operations()
                # for m in op:
                #    print(m)

                # b1 = tf.cast(b1, dtype='float32')
                layer1 = tf.compat.v1.nn.sigmoid(tf.compat.v1.matmul(X, W1) + b1, name='layer1')
                layer1_dropout = tf.compat.v1.nn.dropout(layer1, dropout_keep_prob, name='layer1_dropout')

                noised_sentence = SpellingErrorCorrection.encode_noise(args[0], noise_rate=0.1)
                denoised_sentence = noised_sentence[:]  # will be changed with predict
                for start in range(0, len(noised_sentence) - self.window_size + 1):
                    chars = denoised_sentence[start: start + self.window_size]
                    original_chars = args[0][start: start + self.window_size]
                    _features = [chars]
                    _labels = [original_chars]

                    dataset = DataSet(features=_features, labels=_labels, features_vector=self.features_vector,
                                      labels_vector=self.features_vector)
                    dataset.convert_to_one_hot_vector()
                    try:
                        _y_hat, _cost, _accuracy = sess.run(['y_hat:0', 'cost:0', 'accuracy:0'],
                                                            feed_dict={X: dataset.features, Y: dataset.labels,
                                                                       dropout_keep_prob: dropout_keep_rate})

                        y_hats = [self.features_vector.to_values(_l) for _l in _y_hat]
                        #if _features[0] == y_hats[0]:
                        #    print('same   : "%s"' % (_features[0]))
                        #else:
                        #    print('denoise: "%s" -> "%s"' % (_features[0], y_hats[0]))
                        denoised_sentence = denoised_sentence.replace(_features[0], y_hats[0])
                    except:
                        print('"%s"%s "%s"%s' % (chars, dataset.features.shape, original_chars, dataset.labels.shape))

                    #print(denoised_sentence)

        return denoised_sentence



import pyLDAvis.gensim
import pickle
import gensim

dictionary = gensim.corpora.Dictionary.load('dictionary.gensim')
corpus = pickle.load(open('corpus.pkl', 'rb'))
lda = gensim.models.ldamodel.LdaModel.load('model5.gensim')

lda_display = pyLDAvis.gensim.prepare(lda, corpus, dictionary, sort_topics=False)
#pyLDAvis.display(lda_display)

pyLDAvis.save_html(lda_display, 'vis.html')

from gensim.test.utils import common_corpus

print(str(common_corpus))


from treform.utility import *

from gensim.models import fasttext
from soynlp.hangle import decompose, compose
import re

class Utility:
    def __init__(self):
        name = 'Utility Class'
        self.doublespace_pattern = re.compile('\s+')

    def jamo_sentence(self, sent):
        '''
            초/중/종성이 완전한 한글을 제외한 다른 글자를 제거하며 음절을 초/중/종성으로 분리하는 코드
        :param sent:
        :return:
        '''
        def transform(char):
            if char == ' ':
                return char
            cjj = decompose(char)
            if cjj != None:
                if len(cjj) == 1:
                    return cjj
                cjj_ = ''.join(c if c != ' ' else '-' for c in cjj)

            if cjj == None:
                return ''

            return cjj_

        sent_ = ''.join(transform(char) for char in sent)
        sent_ = self.doublespace_pattern.sub(' ', sent_)
        return sent_

    def decode(self, s):
        '''
            초/중/종성으로 분리된 음절을 다시 원상 복귀하는 코드
        :param s:
        :return:
        '''
        def process(t):
            assert len(t) % 3 == 0
            t_ = t.replace('-', ' ')
            chars = [tuple(t_[3 * i:3 * (i + 1)]) for i in range(len(t_) // 3)]
            recovered = []
            for char in chars:
                try:
                    composed = compose(*char)
                    recovered.append(composed)
                except:
                    pass
            #recovered = [compose(*char) for char in chars]
            recovered = ''.join(recovered)
            return recovered

        return ' '.join(process(t) for t in s.split())

    def decode_sentence(self, sent):
        return ' '.join(self.decode(token) for token in sent.split())

    def cosine_similarity(self, word1, word2, model):
        '''
             초/중/종성으로 분리된 두 음절 사이의 유사도 계산
        :param word1:
        :param word2:
        :param model:
        :return:
        '''
        cjj1 = self.jamo_sentence(word1)
        cjj2 = self.jamo_sentence(word2)
        cos_sim = model.cosine_similarity(cjj1, cjj2)
        return cos_sim

    def most_similar(self, word, model):
        jamo_result = []
        cjj = self.jamo_sentence(word)
        result = model.most_similar(cjj)
        for token in result:
            word = token[0]
            encoded_word = self.decode(word)
            sim = token[1]
            jamo_result.append((encoded_word,sim))

        return jamo_result

    def most_similars(self, model, positives, negatives, topn=10):
        jamo_result = []
        result = model.most_similar(positive=positives,negative=negatives,topn=topn)
        for token in result:
            word = token[0]
            if len(word) > 3:
                encoded_word = self.decode(word)
                sim = token[1]
                jamo_result.append((encoded_word,sim))

        return jamo_result

    def similar_by_word(self, model, word, topn=10):
        jamo_result = []
        result = model.similar_by_word(word, topn=topn)
        for token in result:
            word = token[0]
            if len(word) > 3:
                encoded_word = self.decode(word)
                sim = token[1]
                jamo_result.append((encoded_word, sim))

        return jamo_result


import logging
import os
import sys
import warnings
from os.path import expanduser

from treform.utility.base_util import db_hostname, is_my_pc
from treform.utility.log_util import LogUtil

warnings.simplefilter(action='ignore', category=FutureWarning)  # ignore future warnings

log = None
if log is None:
    if len(sys.argv) == 1:  # by Pycharm or console
        if is_my_pc():  # my pc (pycharm client, mac)
            log = LogUtil.get_logger(None, level=logging.DEBUG, console_mode=True)  # global log
        else:  # gpu pc (batch job, ubuntu)
            log = LogUtil.get_logger(sys.argv[0], level=logging.DEBUG, console_mode=True)  # global log # console_mode=True for jupyter
    else:  # by batch script
        log = LogUtil.get_logger(sys.argv[0], level=logging.INFO, console_mode=False)  # global log

HOME_DIR = 'D:\\python_workspace\\treform'
log.info('HOME_DIR: %s' % HOME_DIR)

PROJECT_DIR = os.path.join(HOME_DIR, '')

DATA_DIR = os.path.join(PROJECT_DIR, 'data')
log.info('DATA_DIR: %s' % DATA_DIR)
if not os.path.exists(DATA_DIR):
    os.mkdir(DATA_DIR)

MODELS_DIR = os.path.join(PROJECT_DIR, 'models')
log.info('MODELS_DIR: %s' % MODELS_DIR)
if not os.path.exists(MODELS_DIR):
    os.mkdir(MODELS_DIR)

#################################################
# tensorboard log dir
#################################################
TENSORBOARD_LOG_DIR = os.path.join(HOME_DIR, 'tensorboard_log')
# log.info('TENSORBOARD_LOG_DIR: %s' % TENSORBOARD_LOG_DIR)
if not os.path.exists(TENSORBOARD_LOG_DIR):
    os.mkdir(TENSORBOARD_LOG_DIR)

#################################################
# mnist
#################################################
MNIST_DIR = os.path.join(HOME_DIR, 'workspace', 'nlp4kor_tensorflow-mnist')
MNIST_DATA_DIR = os.path.join(MNIST_DIR, 'data')
MNIST_CNN_MODEL_DIR = os.path.join(MNIST_DIR, 'models', 'cnn')
MNIST_DAE_MODEL_DIR = os.path.join(MNIST_DIR, 'models', 'dae')

#################################################
# ko.wikipedia.org
#################################################
KO_WIKIPEDIA_ORG_DIR = os.path.join(HOME_DIR, 'treform/spelling')

KO_WIKIPEDIA_ORG_INFO_FILE = os.path.join(KO_WIKIPEDIA_ORG_DIR, 'ko.wikipedia.org.info.txt')
KO_WIKIPEDIA_ORG_URLS_FILE = os.path.join(KO_WIKIPEDIA_ORG_DIR, 'ko.wikipedia.org.urls.txt')
KO_WIKIPEDIA_ORG_CHARACTERS_FILE = os.path.join(KO_WIKIPEDIA_ORG_DIR, 'ko.wikipedia.org.characters')

KO_WIKIPEDIA_ORG_SENTENCES_FILE = os.path.join(KO_WIKIPEDIA_ORG_DIR, 'ko.wikipedia.org.sentences.gz')
KO_WIKIPEDIA_ORG_TRAIN_SENTENCES_FILE = os.path.join(KO_WIKIPEDIA_ORG_DIR, 'ko.wikipedia.org.train.sentences.gz')
KO_WIKIPEDIA_ORG_VALID_SENTENCES_FILE = os.path.join(KO_WIKIPEDIA_ORG_DIR, 'ko.wikipedia.org.valid.sentences.gz')
KO_WIKIPEDIA_ORG_TEST_SENTENCES_FILE = os.path.join(KO_WIKIPEDIA_ORG_DIR, 'ko.wikipedia.org.test.sentences.gz')

KO_WIKIPEDIA_ORG_SPELLING_ERROR_CORRECTION_MODEL_DIR = os.path.join(KO_WIKIPEDIA_ORG_DIR, 'models', 'spelling_error_correction')

#################################################
# ko.wikipedia.org
#################################################
WIKIPEDIA_DIR = 'D:\\python_workspace\\treform\\treform\\spelling'

# text (with string)
#WIKIPEDIA_DATA_DIR = os.path.join(WIKIPEDIA_DIR, 'data')
WIKIPEDIA_DATA_DIR = WIKIPEDIA_DIR
if not os.path.exists(WIKIPEDIA_DATA_DIR):
    os.mkdir(WIKIPEDIA_DATA_DIR)

# info
WIKIPEDIA_INFO_FILE = os.path.join(WIKIPEDIA_DATA_DIR, 'ko.wikipedia.org.info.txt')
WIKIPEDIA_URLS_FILE = os.path.join(WIKIPEDIA_DATA_DIR, 'ko.wikipedia.org.urls.txt')

WIKIPEDIA_CHARACTERS_FILE = os.path.join(WIKIPEDIA_DATA_DIR, 'ko.wikipedia.org.characters')
WIKIPEDIA_SENTENCES_FILE = os.path.join(WIKIPEDIA_DATA_DIR, 'ko.wikipedia.org.sentences.gz')

WIKIPEDIA_TRAIN_SENTENCES_FILE = os.path.join(WIKIPEDIA_DATA_DIR, 'ko.wikipedia.org.train.sentences.gz')
WIKIPEDIA_VALID_SENTENCES_FILE = os.path.join(WIKIPEDIA_DATA_DIR, 'ko.wikipedia.org.valid.sentences.gz')
WIKIPEDIA_TEST_SENTENCES_FILE = os.path.join(WIKIPEDIA_DATA_DIR, 'ko.wikipedia.org.test.sentences.gz')

# csv (with character id)
WIKIPEDIA_TRAIN_FILE = os.path.join(WIKIPEDIA_DATA_DIR, 'ko.wikipedia.org.train.sentences.cid.gz')
WIKIPEDIA_VALID_FILE = os.path.join(WIKIPEDIA_DATA_DIR, 'ko.wikipedia.org.valid.sentences.cid.gz')
WIKIPEDIA_TEST_FILE = os.path.join(WIKIPEDIA_DATA_DIR, 'ko.wikipedia.org.test.sentences.cid.gz')

WIKIPEDIA_NE_FILE = os.path.join(WIKIPEDIA_DATA_DIR, 'ko.wikipedia.org.sentences.ne.gz')


import benepar
import nltk
from nltk import tree

class BaseSyntacticParser:
    IN_TYPE = [list, tuple]
    OUT_TYPE = [list, str]

class BeneparSyntacticParser(BaseSyntacticParser):
    def __init__(self):
        try:
            self.parser = benepar.Parser("benepar_ko2")
        except LookupError:
            benepar.download('benepar_ko2')
            print('error')
        if self.parser is None:
            self.parser = benepar.Parser("benepar_ko2")

    def __call__(self, *args, **kwargs):
        words = []
        tags = []
        for tok in args[0]:
            words.append(tok[0])
            tags.append(tok[1])

        input_sentence = benepar.InputSentence(
            words=words,
            tags=tags,
        )
        tree = self.parser.parse(input_sentence)
        #print(tree)
        return tree._pformat_flat("", "()", False)

class NLTKRegexSyntacticParser(BaseSyntacticParser):
    def __init__(self):
        # Define a chunk grammar, or chunking rules, then chunk
        grammar = """
        NP: {<N.*>*<Suffix>?}   # Noun phrase
        VP: {<V.*>*}            # Verb phrase
        AP: {<A.*>*}            # Adjective phrase
        """
        self.parser = nltk.RegexpParser(grammar)

    def __call__(self, *args, **kwargs):
        chunks = self.parser.parse(*args)
        return chunks._pformat_flat("", "()", False)


import math
import numpy as np

def tf_eccd(corpus,package):

	labelset = package ["labelset"]
	weights = package ["weights"]
	doclist = package ["doclist"]
	doclist = dict(doclist)
	dictlist = {}
	doclen = {}
	n = len(corpus)
	worddict = {}

	for i in corpus:
		labell = i["label"]
		docl = i["document"]
		if labell not in doclen:
			doclen[labell] = 0
		doclen[labell] += i["length"]
		for j in i["split_sentence"].split():
			if labell not in dictlist:
				dictlist[labell] = {}
			if j not in dictlist[labell]:
				dictlist[labell][j] = 1
			else:
				dictlist[labell][j] += 1

			if labell not in doclist:
				doclist[labell] = {}
			if j not in doclist[labell]:
				doclist[labell][j] = set()
			doclist[labell][j].add(docl)

	entropy = {}
	tf = {}
	for labell in labelset:
		tf[labell] = {}
		for word in doclist[labell]:
			if word not in worddict:
				worddict[word] = 0
			worddict[word] += len(doclist[labell][word])
			total = sum([dictlist[x][word]  for x in dictlist if word in dictlist[x]])
			tf[labell][word] = dictlist[labell][word]*1.0/total

	Emax = -999
	for labell in labelset:
		for word in doclist[labell]:
			if word not in entropy:
				entropy[word] = -sum([ tf[x][word]*1.0*math.log(tf[x][word],2) for x in labelset if word in tf[x] ])
				Emax = max(entropy[word],Emax)

	for labell in labelset:
		weights[labell] = {}
		for word in doclist[labell]:
			a_b = worddict[word]
			a = len(doclist[labell][word])
			b = a_b - a
			c_d = sum([(worddict[x]) for x in worddict.keys() if x!=word])
			c = sum([len(doclist[labell][x]) for x in (doclist[labell]) if x!=word])
			d = c_d - c
			if ((a*d) - (b*c)) != 0 and Emax != 0:
				weights[labell][word] = ((a*d-b*c)*1.0/((a+c)*(b+d)))*(Emax - entropy[word])*1.0/Emax

	tf_eccd = {}
	for labell in labelset:
		tf_eccd[labell] = {}
		for word in dictlist[labell]:
			tf_eccd[labell][word] = dictlist[labell][word]*1.0 /  (doclen[labell]*1.0)
			if word in weights[labell]:
				tf_eccd[labell][word] *= weights[labell][word]

	package ["labelset"] = labelset
	package ["weights"] = weights
	package ["doclist"] = doclist

	return tf_eccd



from pickle import load
from numpy import array
from keras.preprocessing.text import Tokenizer
from keras.preprocessing.sequence import pad_sequences
from keras.models import load_model

class sentimentPredictor:
    def __init__(self):
        name = 'sentimentPredictor'

    # load a clean dataset
    def load_dataset(self, filename):
        return load(open(filename, 'rb'))

    # fit a tokenizer
    def create_tokenizer(self, lines):
        tokenizer = Tokenizer()
        tokenizer.fit_on_texts(lines)
        return tokenizer


    # calculate the maximum document length
    def max_length(self, lines):
        return max([len(s.split()) for s in lines])


    # encode a list of lines
    def encode_text(self, tokenizer, lines, length):
        # integer encode
        encoded = tokenizer.texts_to_sequences(lines)
        # pad encoded sequences
        padded = pad_sequences(encoded, maxlen=length, padding='post')
        return padded

if __name__ == '__main__':
    # load datasets
    trainLines, trainLabels = sentimentPredictor().load_dataset('train.pkl')
    testLines, testLabels = sentimentPredictor().load_dataset('test.pkl')

    # create tokenizer
    tokenizer = sentimentPredictor().create_tokenizer(trainLines)
    # calculate max document length
    length = sentimentPredictor().max_length(trainLines)
    # calculate vocabulary size
    vocab_size = len(tokenizer.word_index) + 1
    print('Max document length: %d' % length)
    print('Vocabulary size: %d' % vocab_size)
    # encode data
    trainX = sentimentPredictor().encode_text(tokenizer, trainLines, length)
    testX = sentimentPredictor().encode_text(tokenizer, testLines, length)
    print(trainX.shape, testX.shape)

    # load the model
    model = load_model('model.h5')

    # evaluate model on training dataset
    loss, acc = model.evaluate([trainX, trainX, trainX], array(trainLabels), verbose=0)
    print('Train Accuracy: %f' % (acc * 100))

    # evaluate model on test dataset dataset
    loss, acc = model.evaluate([testX, testX, testX], array(testLabels), verbose=0)
    print('Test Accuracy: %f' % (acc * 100))

    """
   tokenizer1 = sentimentPredictor().create_tokenizer(['this is a good news'])
   # calculate max document length
   length1 = sentimentPredictor().max_length(['this is a good news'])
   # calculate vocabulary size
   vocab_size1 = len(tokenizer1.word_index) + 1
   print('Max document length: %d' % length1)
   print('Vocabulary size: %d' % vocab_size1)
   # encode data
   test = sentimentPredictor().encode_text(tokenizer1, ['this is a good news'], length1)
   """

    # make a prediction
    test_predict = model.predict([testX, testX, testX], verbose=1, batch_size=1)

    print(test_predict)



import pickle
from os import listdir
from nltk.corpus import stopwords
from pickle import dump
from string import punctuation
import os, sys
from stat import *
import treform as ptm
import io

class sentimentDataManager:
    def __init__(self):
        name = 'sentimentDataManager'

    # load doc into memory
    def load_doc(self, filename):
        # open the file as read only
        file = open(filename, 'r')
        # read all text
        text = file.read()
        # close the file
        file.close()
        return text

    # save a dataset to file
    def save_dataset(self, dataset, filename):
        dump(dataset, open(filename, 'wb'))
        print('Saved: %s' % filename)

    def load_dataset(self, filename):
        # load the model from disk
        loaded_model = pickle.load(open(filename, 'rb'))
        return loaded_model

if __name__ == '__main__':

    _negative_docs = ptm.CorpusFromDirectory('txt_sentoken/neg', True)
    _positive_docs = ptm.CorpusFromDirectory('txt_sentoken/pos', True)

    pipeline = ptm.Pipeline(ptm.splitter.NLTK(),
                            ptm.tokenizer.Word(),
                            ptm.helper.StopwordFilter(file='../stopwordsEng.txt'),
                            ptm.stemmer.Porter())
    _neg_result = pipeline.processCorpus(_negative_docs)
    _pos_result = pipeline.processCorpus(_positive_docs)
    print('== Splitting Sentence + Tokenizing + Stopwords Removal + Stemming : Porter ==')
    print(_neg_result)
    print()

    negative_docs = list()
    for doc in _neg_result:
        new_doc = []
        for sent in doc:
            for _str in sent:
                if len(_str) > 0:
                    new_doc.append(_str)
        negative_docs.append(' '.join(new_doc))

    positive_docs = list()
    for doc in _pos_result:
        new_doc = []
        for sent in doc:
            for _str in sent:
                if len(_str) > 0:
                    new_doc.append(_str)
        positive_docs.append(' '.join(new_doc))


    # load all training reviews
    trainX = negative_docs + positive_docs
    print("TRAIN X " + str(len(negative_docs)))

    trainy = [0 for _ in range(900)] + [1 for _ in range(900)]
    sentimentDataManager().save_dataset([trainX, trainy], 'train.pkl')

    # load all test reviews
    _negative_docs = ptm.CorpusFromDirectory('txt_sentoken/neg', False)
    _positive_docs = ptm.CorpusFromDirectory('txt_sentoken/pos', False)

    pipeline = ptm.Pipeline(ptm.splitter.NLTK(),
                            ptm.tokenizer.Word(),
                            ptm.helper.StopwordFilter(file='../stopwordsEng.txt'),
                            ptm.stemmer.Porter())
    _neg_result = pipeline.processCorpus(_negative_docs)
    _pos_result = pipeline.processCorpus(_positive_docs)
    print('== Splitting Sentence + Tokenizing + Stopwords Removal + Stemming : Porter ==')
    print(_neg_result)
    print()

    negative_docs = list()
    for doc in _neg_result:
        new_doc = []
        for sent in doc:
            for _str in sent:
                if len(_str) > 0:
                    new_doc.append(_str)
        negative_docs.append(' '.join(new_doc))

    positive_docs = list()
    for doc in _pos_result:
        new_doc = []
        for sent in doc:
            for _str in sent:
                if len(_str) > 0:
                    new_doc.append(_str)
        positive_docs.append(' '.join(new_doc))

    testX = negative_docs + positive_docs
    testY = [0 for _ in range(100)] + [1 for _ in range(100)]
    sentimentDataManager().save_dataset([testX, testY], 'test.pkl')



import datetime
import logging.handlers
import os
import re
import time
from logging.handlers import BaseRotatingHandler
from random import randint

# sibling module than handles all the ugly platform-specific details of file locking
from portalocker import lock, unlock, LOCK_EX

__version__ = '0.0.1'
__author__ = "yorks"


# noinspection PyUnusedLocal
class MultProcTimedRotatingFileHandler(BaseRotatingHandler):
    """
    Handler for logging to a file, rotating the log file at certain timed.
    https://github.com/yorks/mpfhandler
    """

    def __init__(self, filename, when='h', interval=1, backupCount=0, encoding=None, delay=False, utc=False,
                 debug=False):
        """
            * interval, backupCount is not working!!! *

        Just Copied from logging.handlers.TimedRotatingFileHandler

        # a rollover occurs.  Current 'when' events supported:
        # S - Seconds
        # M - Minutes
        # H - Hours
        # D - Days
        # midnight - roll over at midnight
        # W{0-6} - roll over on a certain day; 0 - Monday
        #
        # Case of the 'when' specifier is not important; lower or upper case
        # will work.

        """
        BaseRotatingHandler.__init__(self, filename, 'a', encoding, delay)
        self.when = when.upper()
        self.backupCount = backupCount
        self.utc = utc
        self.debug = debug
        self.mylogfile = "%s.%08d" % ('/tmp/mptfhanldler', randint(0, 99999999))

        self.interval = 1  # datetime timedelta only have, days, seconds, microseconds

        if self.when == 'S':
            # self.interval = 1 # one second
            self.suffix = "%Y-%m-%d_%H-%M-%S"
            self.extMatch = r"^\d{4}-\d{2}-\d{2}_\d{2}-\d{2}-\d{2}$"
        elif self.when == 'M':
            self.interval = 60  # one minute
            self.suffix = "%Y-%m-%d_%H-%M"
            self.extMatch = r"^\d{4}-\d{2}-\d{2}_\d{2}-\d{2}$"
        elif self.when == 'H':
            self.interval = 60 * 60  # one hour
            self.suffix = "%Y-%m-%d_%H"
            self.extMatch = r"^\d{4}-\d{2}-\d{2}_\d{2}$"
        elif self.when == 'D' or self.when == 'MIDNIGHT':
            # self.interval = 60 * 60 * 24 # one day
            self.suffix = "%Y-%m-%d"
            self.extMatch = r"^\d{4}-\d{2}-\d{2}$"
            self.when = 'D'  # MIDNIGHT is day, use day only
        elif self.when.startswith('W'):
            # self.interval = 60 * 60 * 24 * 7 # one week
            if len(self.when) != 2:
                raise ValueError("You must specify a day for weekly rollover from 0 to 6 (0 is Monday): %s" % self.when)
            if self.when[1] < '0' or self.when[1] > '6':
                raise ValueError("Invalid day specified for weekly rollover: %s" % self.when)
            self.dayOfWeek = int(self.when[1])
            self.suffix = "%Y-%m-%d"
            self.extMatch = r"^\d{4}-\d{2}-\d{2}$"
        else:
            raise ValueError("Invalid rollover interval specified: %s" % self.when)

        self.extMatch = re.compile(self.extMatch)
        # self.interval = self.interval * interval # multiply by units requested
        # self.interval = self.interval * 1  # interval arg is not working

        # lock file, contain next rollover timestamp
        self.stream_lock = None
        self.lock_file = self._getLockFile()

        # read from conf first for inherit the first process
        # if it is the first process, please remove the lock file by hand first
        self.nextRolloverTime = self.getNextRolloverTime()
        if not self.nextRolloverTime:
            self.nextRolloverTime = self.computerNextRolloverTime()
            self.saveNextRolloverTime()

    def _log2mylog(self, msg):
        time_str = time.strftime('%Y-%m-%d %H:%M:%S', time.localtime(time.time()))
        msg = str(msg)
        content = "%s [%s]\n" % (time_str, msg)
        fa = open(self.mylogfile, 'a')
        fa.write(content)
        fa.close()

    def _getLockFile(self):
        lock_file = os.path.join(os.path.dirname(self.baseFilename),
                                 '.' + os.path.basename(self.baseFilename) + '.lock')
        return lock_file

    def _openLockFile(self):
        lock_file = self._getLockFile()
        self.stream_lock = open(lock_file, 'w')

    def computerNextRolloverTime(self):
        """ Work out the next rollover time. """
        nextTime = None
        currentDateTime = datetime.datetime.now()
        if self.utc:
            currentDateTime = datetime.datetime.utcnow()

        if self.when == 'D':
            nextDateTime = currentDateTime + datetime.timedelta(days=self.interval)
            nextDate = nextDateTime.date()
            nextTime = int(time.mktime(nextDate.timetuple()))
        elif self.when.startswith('W'):
            days = 0
            currentWeekDay = currentDateTime.weekday()
            if currentWeekDay == self.dayOfWeek:
                days = (self.interval + 7)
            elif currentWeekDay < self.dayOfWeek:
                days = self.dayOfWeek - currentWeekDay
            else:
                days = 6 - currentWeekDay + self.dayOfWeek + 1
            nextDateTime = currentDateTime + datetime.timedelta(days=days)
            nextDate = nextDateTime.date()
            nextTime = int(time.mktime(nextDate.timetuple()))
        else:
            tmpNextDateTime = currentDateTime + datetime.timedelta(seconds=self.interval)
            nextDateTime = tmpNextDateTime.replace(microsecond=0)
            if self.when == 'H':
                nextDateTime = tmpNextDateTime.replace(minute=0, second=0, microsecond=0)
            elif self.when == 'M':
                nextDateTime = tmpNextDateTime.replace(second=0, microsecond=0)

            nextTime = int(time.mktime(nextDateTime.timetuple()))
        return nextTime

    def getNextRolloverTime(self):
        """ get next rollover time stamp from lock file """
        try:
            fp = open(self.lock_file, 'r')
            c = fp.read()
            fp.close()
            return int(c)
        except:
            return False

    def saveNextRolloverTime(self):
        """ save the nextRolloverTimestamp to lock file

            this is a flag for avoid multiple processes to rotate
            the log file again at the same rollovertime.
        """
        if not self.nextRolloverTime:
            return 0
        content = "%d" % self.nextRolloverTime

        if not self.stream_lock:
            self._openLockFile()
        lock(self.stream_lock, LOCK_EX)
        try:
            self.stream_lock.seek(0)
            self.stream_lock.write(content)
            self.stream_lock.flush()
        except:
            if self.debug:
                self._log2mylog('saveNextRT exception!!!')
            pass
        finally:
            unlock(self.stream_lock)
        if self.debug:
            self._log2mylog('saveNextRT:%s' % content)

    def acquire(self):
        """ Acquire thread and file locks.

            Copid from ConcurrentRotatingFileHandler
        """
        # handle thread lock
        BaseRotatingHandler.acquire(self)
        # Issue a file lock.  (This is inefficient for multiple active threads
        # within a single process. But if you're worried about high-performance,
        # you probably aren't using this log handler.)
        if self.stream_lock:
            # If stream_lock=None, then assume close() was called or something
            # else weird and ignore all file-level locks.
            if self.stream_lock.closed:
                # Daemonization can close all open file descriptors, see
                # https://bugzilla.redhat.com/show_bug.cgi?id=952929
                # Try opening the lock file again.  Should we warn() here?!?
                try:
                    self._openLockFile()
                except Exception:
                    # Don't try to open the stream lock again
                    self.stream_lock = None
                    return
            lock(self.stream_lock, LOCK_EX)
            # Stream will be opened as part by FileHandler.emit()

    def release(self):
        """ Release file and thread locks.
        """
        try:
            if self.stream_lock and not self.stream_lock.closed:
                unlock(self.stream_lock)
        except Exception:
            pass
        finally:
            # release thread lock
            BaseRotatingHandler.release(self)

    def _close_stream(self):
        """ Close the log file stream """
        if self.stream:
            try:
                if not self.stream.closed:
                    self.stream.flush()
                    self.stream.close()
            finally:
                self.stream = None

    def _close_stream_lock(self):
        """ Close the lock file stream """
        if self.stream_lock:
            try:
                if not self.stream_lock.closed:
                    self.stream_lock.flush()
                    self.stream_lock.close()
            finally:
                self.stream_lock = None

    def close(self):
        """
        Close log stream and stream_lock. """
        try:
            self._close_stream()
            self._close_stream_lock()
        finally:
            self.stream = None
            self.stream_lock = None

    def shouldRollover(self, record):
        """
        Determine if rollover should occur.

        record is not used, as we are just comparing times, but it is needed so
        the method signatures are the same

        Copied from std lib
        """
        t = int(time.time())
        if t >= self.nextRolloverTime:
            return 1
        # print "No need to rollover: %d, %d" % (t, self.rolloverAt)
        return 0

    def doRollover(self):
        """ Do a rollover,

            0. close stream, stream_lock file handle
            1. get lock
            2. mv log log.$date
            3. setting up nextRolloverTime
            4. relese lock
        """
        if self.debug:
            self._log2mylog('do Rollover')
        self._close_stream()
        self.acquire()
        try:
            fileNextRolloverTime = self.getNextRolloverTime()
            if not fileNextRolloverTime:
                if self.debug:
                    self._log2mylog('getNextRolloverTime False, skip rotate!')
                self.release()
                return 0
            # avoid other process do rollover again.
            if self.nextRolloverTime < fileNextRolloverTime:
                self.nextRolloverTime = fileNextRolloverTime
                if self.debug:
                    self._log2mylog('already rotated, skip this proc to rotate!')
                self.release()
                return 0
        except Exception as e:
            pass
        # because log is older then self.nextRolloverTime,
        # we need the old log rename to old filename
        #   donot use time.time()-1,
        #   for between last rollover and nextRolloverTime*N may have none log to record.
        time_tuple = time.localtime(self.nextRolloverTime - 1)
        dfn = self.baseFilename + "." + time.strftime(self.suffix, time_tuple)
        # rename
        if os.path.exists(dfn):
            bakname = dfn + ".bak"
            while os.path.exists(bakname):
                bakname = "%s.%08d" % (bakname, randint(0, 99999999))
            try:
                os.rename(dfn, bakname)
            except:
                pass
        if os.path.exists(self.baseFilename):
            try:
                if self.debug:
                    self._log2mylog('rename %s to %s' % (self.baseFilename, dfn))
                os.rename(self.baseFilename, dfn)
            except:
                pass
        # set new nextRolloverTime
        self.nextRolloverTime = self.computerNextRolloverTime()
        self.saveNextRolloverTime()

        if not self.delay:
            self.stream = self._open()
        self.release()


logging.handlers.MultProcTimedRotatingFileHandler = MultProcTimedRotatingFileHandler


import json
from argparse import ArgumentParser
from typing import List, Tuple

import tensorflow as tf


class SpacingModel(tf.keras.Model):
    def __init__(
            self,
            vocab_size: int,
            hidden_size: int,
            num_classes: int = 3,
            conv_activation: str = "relu",
            dense_activation: str = "relu",
            conv_kernel_and_filter_sizes: List[Tuple[int, int]] = [
                (2, 8),
                (3, 8),
                (4, 8),
                (5, 8),
            ],
            dropout_rate: float = 0.3,
    ):
        super().__init__()

        self.embeddings = tf.keras.layers.Embedding(vocab_size, hidden_size)
        self.convs = [
            tf.keras.layers.Conv1D(
                filter_size,
                kernel_size,
                padding="same",
                activation=conv_activation,
            )
            for kernel_size, filter_size in conv_kernel_and_filter_sizes
        ]
        self.pools = [
            tf.keras.layers.MaxPooling1D(pool_size=filter_size, data_format="channels_first")
            for _, filter_size in conv_kernel_and_filter_sizes
        ]
        self.dropout1 = tf.keras.layers.Dropout(rate=dropout_rate)
        self.output_dense1 = tf.keras.layers.Dense(hidden_size, activation=dense_activation)
        self.dropout2 = tf.keras.layers.Dropout(rate=dropout_rate)
        self.output_dense2 = tf.keras.layers.Dense(num_classes)

    def call(self, input_tensor):
        """
        input_tensor: Tokenized Sequences, Shape: (Batch Size, Sequence Length)
        """

        # embeddings: (Batch Size, Sequence Length, Hidden Size)
        embeddings = self.embeddings(input_tensor)
        # features: (Batch Size, Sequence Length, sum(#filters))
        features = self.dropout1(
            tf.concat([pool(conv(embeddings)) for conv, pool in zip(self.convs, self.pools)], axis=-1)
        )
        # projected: (Batch Size, Sequence Length, Hidden Size)
        projected = self.dropout2(self.output_dense1(features))
        # (Batch Size, Sequence Length, 2)
        return self.output_dense2(projected)


def string_to_example(
        vocab_table: tf.lookup.StaticHashTable,
        encoding: str = "UTF-8",
        max_length: int = 256,
        delete_prob: float = 0.5,
        add_prob: float = 0.15,
):
    @tf.function
    def _inner(tensors: tf.Tensor):
        bytes_array = tf.strings.unicode_split(tf.strings.regex_replace(tensors, " +", " "), encoding)
        space_positions = bytes_array == " "
        sequence_length = tf.shape(space_positions)[0]

        while_condition = lambda i, *_: i < sequence_length

        def while_body(i, strings, labels):
            # 다음 char가 space가 아니고, 문장 끝이 아닐 때 add_prob의 확률로 space 추가
            # 이번 char가 space일 때
            is_next_char_space = tf.cond(i < sequence_length - 1, lambda: bytes_array[i + 1] == " ", lambda: False)

            state = tf.cond(
                is_next_char_space,
                lambda: tf.cond(tf.random.uniform([]) < delete_prob, lambda: 2, lambda: 0),
                lambda: tf.cond(bytes_array[i] != " " and tf.random.uniform([]) < add_prob, lambda: 1, lambda: 0),
            )
            # 0: 그대로 진행
            # 1: 다음 인덱스에 space 추가
            # 2: 다음 space 삭제
            strings = tf.cond(
                state != 1,
                lambda: tf.concat([strings, [bytes_array[i]]], axis=0),
                lambda: tf.concat([strings, [bytes_array[i], " "]], axis=0),
            )
            # label 0: 변화 x
            # label 1: 다음 인덱스에 space 추가
            # label 2: 현재 space 삭제
            labels = tf.cond(
                state == 0,
                lambda: tf.concat([labels, [0]], axis=0),
                lambda: tf.cond(
                    state == 1,
                    lambda: tf.concat([labels, [0, 2]], axis=0),
                    lambda: tf.concat([labels, [1]], axis=0),
                ),
            )
            i += tf.cond(state == 2, lambda: 2, lambda: 1)

            return (i, strings, labels)

        i, strings, labels = tf.while_loop(
            while_condition,
            while_body,
            (
                tf.constant(0),
                tf.constant([], dtype=tf.string),
                tf.constant([], dtype=tf.int32),
            ),
            shape_invariants=(tf.TensorShape([]), tf.TensorShape([None]), tf.TensorShape([None])),
        )

        strings = vocab_table.lookup(tf.concat([["<s>"], strings, ["</s>"]], axis=0))
        labels = tf.concat([[0], labels, [0]], axis=0)

        strings = tf.cond(tf.shape(strings)[0] > max_length, lambda: strings[:max_length], lambda: strings)
        labels = tf.cond(tf.shape(labels)[0] > max_length, lambda: labels[:max_length], lambda: labels)

        length_to_pad = max_length - tf.shape(strings)[0]
        strings = tf.pad(strings, [[0, length_to_pad]])
        labels = tf.pad(labels, [[0, length_to_pad]], constant_values=-1)

        return (strings, labels)

    return _inner


def sparse_categorical_crossentropy_with_ignore(y_true, y_pred, from_logits=False, axis=-1, ignore_id=-1):
    positions = tf.where(y_true != ignore_id)

    y_true = tf.gather_nd(y_true, positions)
    y_pred = tf.gather_nd(y_pred, positions)

    return tf.keras.losses.sparse_categorical_crossentropy(y_true, y_pred, from_logits=from_logits, axis=axis)


def sparse_categorical_accuracy_with_ignore(y_true, y_pred, ignore_id=-1):
    positions = tf.where(y_true != ignore_id)

    y_true = tf.gather_nd(y_true, positions)
    y_pred = tf.gather_nd(y_pred, positions)

    return tf.keras.metrics.sparse_categorical_accuracy(y_true, y_pred)


class SparseCategoricalCrossentropyWithIgnore(tf.python.keras.losses.LossFunctionWrapper):
    def __init__(
            self,
            from_logits=False,
            reduction=tf.keras.losses.Reduction.AUTO,
            ignore_id=-1,
            name="sparse_categorical_crossentropy_with_ignore",
    ):
        super(SparseCategoricalCrossentropyWithIgnore, self).__init__(
            sparse_categorical_crossentropy_with_ignore,
            name=name,
            reduction=reduction,
            ignore_id=ignore_id,
            from_logits=from_logits,
        )


def train(model_path='./models/checkpoint-{epoch}.ckpt',
          log_dir='./logs',
          train_file='',
          dev_file='',
          training_config='',
          char_file=''):
    # args = parser.parse_args()
    with open(training_config) as f:
        config = json.load(f)

    with open(char_file) as f:
        content = f.read()
        keys = ["<pad>", "<s>", "</s>", "<unk>"] + list(content)
        values = list(range(len(keys)))

        vocab_initializer = tf.lookup.KeyValueTensorInitializer(keys, values, key_dtype=tf.string, value_dtype=tf.int32)
        vocab_table = tf.lookup.StaticHashTable(vocab_initializer, default_value=3)

    train_dataset = (
        tf.data.TextLineDataset(tf.constant([train_file]))
            .shuffle(25000000)
            .map(
            string_to_example(vocab_table),
            num_parallel_calls=tf.data.experimental.AUTOTUNE,
        )
            .batch(config["train_batch_size"])
    )
    dev_dataset = (
        tf.data.TextLineDataset(tf.constant([dev_file]))
            .shuffle(100000)
            .map(
            string_to_example(vocab_table),
            num_parallel_calls=tf.data.experimental.AUTOTUNE,
        )
            .batch(config["val_batch_size"])
            .take(4)
    )

    model = SpacingModel(
        config["vocab_size"],
        config["hidden_size"],
        conv_activation=config["conv_activation"],
        dense_activation=config["dense_activation"],
        conv_kernel_and_filter_sizes=config["conv_kernel_and_filter_sizes"],
        dropout_rate=config["dropout_rate"],
    )
    model.compile(
        optimizer=tf.optimizers.Adam(learning_rate=config["learning_rate"]),
        loss=SparseCategoricalCrossentropyWithIgnore(from_logits=True, ignore_id=-1),
        metrics=[sparse_categorical_accuracy_with_ignore],
    )
    model.fit(
        train_dataset,
        epochs=config["epochs"],
        validation_data=dev_dataset,
        steps_per_epoch=400,
        callbacks=[
            tf.keras.callbacks.ModelCheckpoint(filepath=model_path),
            tf.keras.callbacks.TensorBoard(log_dir=log_dir),
            tf.keras.callbacks.ReduceLROnPlateau(patience=2, verbose=1),
        ],
    )

if __name__ == "__main__":
    parser = ArgumentParser()
    parser.add_argument("--train-file", type=str, required=True)
    parser.add_argument("--dev-file", type=str, required=True)
    parser.add_argument("--training-config", type=str, required=True)
    parser.add_argument("--char-file", type=str, required=True)
    parser.add_argument("--model-path", type=str, required=True)
    parser.add_argument("--log-dir", type=str, required=True)
    args = parser.parse_args()

    model_path = args.model_path
    log_dir = args.log_dir
    train_file = args.train_file
    dev_file = args.dev_file
    training_config = args.training_config
    char_file = args.char_file

    train(model_path=model_path, log_dir=log_dir,
          train_file=train_file, dev_file=dev_file,
          training_config=training_config, char_file=char_file)



import sys
from collections import Counter

import matplotlib
import sklearn
import tomotopy as tp
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
from matplotlib.ticker import FuncFormatter
from wordcloud import WordCloud, STOPWORDS
import matplotlib.colors as mcolors
from matplotlib.patches import Rectangle
import matplotlib as mpl
import platform
import matplotlib.font_manager as fm

class pyTextMinerTopicModel:
    def __init__(self):
        self.name="Topic Model"

    def format_topics_sentences(self, topic_number=20, mdl=None):
        pd.set_option('display.max_columns', None)
        # Init output
        sent_topics_df = pd.DataFrame()

        # we need to make column consistent
        matrix = []
        docs = []
        for d in mdl.docs:
            doc = ''
            for word in d.get_words():
                doc += word[0] + " "
            doc = doc.strip()
            docs.append(doc)

            row = d.get_topics(top_n=topic_number)
            row = sorted(row, key=lambda x: (x[1]), reverse=True)
            # Get the Dominant topic, Perc Contribution and Keywords for each document
            for j, (topic_num, prop_topic) in enumerate(row):
                if j == 0:  # => dominant topic
                    topic_keywords = ", ".join([word for word, prob in mdl.get_topic_words(topic_num)])
                    sent_topics_df = sent_topics_df.append(
                        pd.Series([int(topic_num), round(prop_topic, 4), topic_keywords]), ignore_index=True)
                else:
                    break

            a_row = [0.0 for i in range(topic_number)]
            for tup in row:
                for x in range(topic_number):
                    if tup[0] == x:
                        a_row[x] = tup[1]
            #print(str(a_row))
            matrix.append(a_row)

        sent_topics_df.columns = ['Dominant_Topic', 'Perc_Contribution', 'Topic_Keywords']

        # Add original text to the end of the output
        contents = pd.Series(docs)
        sent_topics_df = pd.concat([sent_topics_df, contents], axis=1)

        return(sent_topics_df, matrix)

    def distribution_document_word_count(self, sent_topics_df, df_dominant_topic):
        # The most representative sentence for each topic
        # Display setting to show more characters in column
        pd.options.display.max_colwidth = 100

        sent_topics_sorteddf_mallet = pd.DataFrame()
        sent_topics_outdf_grpd = sent_topics_df.groupby('Dominant_Topic')

        for i, grp in sent_topics_outdf_grpd:
            sent_topics_sorteddf_mallet = pd.concat([sent_topics_sorteddf_mallet,
                                                     grp.sort_values(['Perc_Contribution'], ascending=False).head(1)],
                                                    axis=0)

        # Reset Index
        sent_topics_sorteddf_mallet.reset_index(drop=True, inplace=True)

        # Format
        sent_topics_sorteddf_mallet.columns = ['Topic_Num', "Topic_Perc_Contrib", "Keywords", "Representative Text"]

        # Show
        print(sent_topics_sorteddf_mallet.head(10))

        # Frequency Distribution of Word Counts in Documents
        # When working with a large number of documents, you want to know how big the documents are as a whole and by topic.
        # Let’s plot the document word counts distribution.
        doc_lens = [len(d) for d in df_dominant_topic.Text]

        # Plot
        plt.figure(figsize=(16, 7), dpi=160)
        plt.hist(doc_lens, bins=1000, color='navy')
        plt.text(750, 100, "Mean   : " + str(round(np.mean(doc_lens))))
        plt.text(750, 90, "Median : " + str(round(np.median(doc_lens))))
        plt.text(750, 80, "Stdev   : " + str(round(np.std(doc_lens))))
        plt.text(750, 70, "1%ile    : " + str(round(np.quantile(doc_lens, q=0.01))))
        plt.text(750, 60, "99%ile  : " + str(round(np.quantile(doc_lens, q=0.99))))

        plt.gca().set(xlim=(0, 200), ylabel='Number of Documents', xlabel='Document Word Count')
        plt.gca().set_ylim([0,200])
        plt.tick_params(size=16)
        plt.xticks(np.linspace(0, 200, 9))
        plt.title('Distribution of Document Word Counts', fontdict=dict(size=22))
        plt.show()

    def distribution_word_count_by_dominant_topic(self, df_dominant_topic):
        import seaborn as sns
        import matplotlib.colors as mcolors
        cols = [color for name, color in mcolors.TABLEAU_COLORS.items()]  # more colors: 'mcolors.XKCD_COLORS'

        fig, axes = plt.subplots(2, 2, figsize=(16, 14), dpi=160, sharex=True, sharey=True)

        for i, ax in enumerate(axes.flatten()):
            df_dominant_topic_sub = df_dominant_topic.loc[df_dominant_topic.Dominant_Topic == i, :]
            doc_lens = [len(d) for d in df_dominant_topic_sub.Text]
            ax.hist(doc_lens, bins=1000, color=cols[i])
            ax.tick_params(axis='y', labelcolor=cols[i], color=cols[i])
            sns.kdeplot(doc_lens, color="black", shade=False, ax=ax.twinx())
            ax.set(xlim=(0, 1000), xlabel='Document Word Count')
            ax.set_ylabel('Number of Documents', color=cols[i])
            ax.set_title('Topic: ' + str(i), fontdict=dict(size=16, color=cols[i]))

        fig.tight_layout()
        fig.subplots_adjust(top=0.90)
        plt.xticks(np.linspace(0, 1000, 9))
        fig.suptitle('Distribution of Document Word Counts by Dominant Topic', fontsize=22)
        plt.show()

    def word_cloud_by_topic(self, mdl):
        #Word Clouds of Top N Keywords in Each Topic
        #Though you’ve already seen what are the topic keywords in each topic, a word cloud
        #with the size of the words proportional to the weight is a pleasant sight.
        #The coloring of the topics I’ve taken here is followed in the subsequent plots as well.
        cols = [color for name, color in mcolors.TABLEAU_COLORS.items()]  # more colors: 'mcolors.XKCD_COLORS'

        if platform.system() == 'Windows':
            # Window의 경우 폰트 경로
            font_path = 'C:/Windows/Fonts/malgun.ttf'
        elif platform.system() == 'Darwin':
            # for Mac
            font_path = '/Library/Fonts/AppleGothic.ttf'

        cloud = WordCloud(background_color='white',
                          font_path=font_path,
                          collocations=False,
                          width=2500,
                          height=1800,
                          max_words=10,
                          colormap='tab10',
                          color_func=lambda *args, **kwargs: cols[i],
                          prefer_horizontal=1.0)

        fig, axes = plt.subplots(2, 2, figsize=(10, 10), sharex=True, sharey=True)

        for i, ax in enumerate(axes.flatten()):
            fig.add_subplot(ax)
            topic_words={}
            for word, prob in mdl.get_topic_words(i):
                topic_words[word]=prob
            cloud.generate_from_frequencies(topic_words, max_font_size=300)
            plt.gca().imshow(cloud)
            plt.gca().set_title('Topic ' + str(i), fontdict=dict(size=16))
            plt.gca().axis('off')

        plt.subplots_adjust(wspace=0, hspace=0)
        plt.axis('off')
        plt.margins(x=0, y=0)
        plt.tight_layout()
        plt.show()

    def word_count_by_keywords(self, mdl, matrix):
        from collections import Counter
        data_flat = [w for w_list in matrix for w in w_list]
        counter = Counter(data_flat)

        font_path = ''
        if platform.system() == 'Windows':
            # Window의 경우 폰트 경로
            font_path = 'C:/Windows/Fonts/malgun.ttf'
        elif platform.system() == 'Darwin':
            # for Mac
            font_path = '/Library/Fonts/AppleGothic.ttf'

        font_name = fm.FontProperties(fname=font_path).get_name()
        plt.rc('font', family=font_name)
        plt.rc('axes', unicode_minus=False)

        out = []
        for k in range(mdl.k):
            for word, weight in mdl.get_topic_words(k):
                out.append([word, k, weight, counter[word]])

        df = pd.DataFrame(out, columns=['word', 'topic_id', 'importance', 'word_count'])

        # Plot Word Count and Weights of Topic Keywords
        fig, axes = plt.subplots(2, 2, figsize=(16, 10), sharey=True, dpi=160)
        cols = [color for name, color in mcolors.TABLEAU_COLORS.items()]
        for i, ax in enumerate(axes.flatten()):
            ax.bar(x='word', height="word_count", data=df.loc[df.topic_id == i, :], color=cols[i], width=0.5, alpha=0.3,
                   label='Word Count')
            ax_twin = ax.twinx()
            ax_twin.bar(x='word', height="importance", data=df.loc[df.topic_id == i, :], color=cols[i], width=0.2,
                        label='Weights')
            ax.set_ylabel('Word Count', color=cols[i])
            ax_twin.set_ylim(0, 0.030);
            ax.set_ylim(0, 3500)
            ax.set_title('Topic: ' + str(i), color=cols[i], fontsize=16)
            ax.tick_params(axis='y', left=False)
            ax.set_xticklabels(df.loc[df.topic_id == i, 'word'], rotation=30, horizontalalignment='right')
            ax.legend(loc='upper left');
            ax_twin.legend(loc='upper right')

        fig.tight_layout(w_pad=2)
        fig.suptitle('Word Count and Importance of Topic Keywords', fontsize=22, y=1.05)
        plt.show()

    def sentences_chart(self, mdl, start=0, end=13, topic_number=20):
        font_path = ''
        if platform.system() == 'Windows':
            # Window의 경우 폰트 경로
            font_path = 'C:/Windows/Fonts/malgun.ttf'
        elif platform.system() == 'Darwin':
            # for Mac
            font_path = '/Library/Fonts/Arial Unicode.ttf'

        font_name = fm.FontProperties(fname=font_path).get_name()
        plt.rc('font', family=font_name)
        plt.rc('axes', unicode_minus=False)

        from collections import defaultdict
        #corp = corpus[start:end]
        mycolors = [color for name, color in mcolors.TABLEAU_COLORS.items()]

        fig, axes = plt.subplots(end-start, 1, figsize=(20, (end-start)*0.95), dpi=160)
        axes[0].axis('off')
        for i, ax in enumerate(axes):
            if i > 0:
                #topic_percs, wordid_topics, wordid_phivalues = lda_model[corp_cur]
                #word_dominanttopic = [(lda_model.id2word[wd], topic[0]) for wd, topic in wordid_topics]
                for idx in range(start, end):
                    d = mdl.docs[idx]
                    topic_percs = []
                    word_dominanttopic = defaultdict(list)

                    row = d.get_topics(top_n=topic_number)
                    row = sorted(row, key=lambda x: (x[1]), reverse=True)
                    # Get the Dominant topic, Perc Contribution and Keywords for each document
                    d_topic_percs=[]
                    for j, (topic_num, prop_topic) in enumerate(row):
                        if j < 5:  # => dominant topic
                            d_topic_percs.append(prop_topic)
                            k=0
                            for word, prob in mdl.get_topic_words(topic_num):
                                if k < 2:
                                    word_dominanttopic[word].append(topic_num)
                                k+=1
                    topic_percs.append(d_topic_percs)

                    ax.text(0.01, 0.5, "Doc " + str(i-1) + ": ", verticalalignment='center',
                            fontsize=16, color='black', transform=ax.transAxes, fontweight=700)

                    # Draw Rectange
                    topic_percs_sorted = sorted(topic_percs, key=lambda x: (x[1]), reverse=True)
                    ax.add_patch(Rectangle((0.0, 0.05), 0.99, 0.90, fill=None, alpha=1,
                                       color=mycolors[len(topic_percs_sorted[0])], linewidth=2))

                    word_pos = 0.08
                    j = 0
                    for word in word_dominanttopic:
                        topics = word_dominanttopic[word]
                        if j < 3:
                            print(str(word) + " : " + str(topics))
                            ax.text(word_pos, 0.5, word,
                                    horizontalalignment='left',
                                    verticalalignment='center',
                                    fontsize=16, color=mycolors[len(topics)],
                                    transform=ax.transAxes, fontweight=700)
                            word_pos += .009 * len(word)  # to move the word for the next iter
                            ax.axis('off')
                        j+=1
                    ax.text(word_pos, 0.5, '. . .',
                            horizontalalignment='left',
                            verticalalignment='center',
                            fontsize=16, color='black',
                            transform=ax.transAxes)
                print("\n")

        plt.subplots_adjust(wspace=0, hspace=0)
        plt.suptitle('Sentence Topic Coloring for Documents: ' + str(start) + ' to ' + str(end-2), fontsize=22, y=0.95, fontweight=700)
        plt.tight_layout()
        plt.show()

    # Sentence Coloring of N Sentences
    def topics_per_document(self, mdl, start=0, end=1):
        if (end < 1):
            end = len(mdl.docs)

        from collections import defaultdict
        topic_top3words = defaultdict(list)

        dominant_topics = []
        topic_percentages = []
        i=0
        for idx in range(start, end):
            d = mdl.docs[idx]
            row = d.get_topics(top_n=topic_number)
            row = sorted(row, key=lambda x: (x[1]), reverse=True)
            # Get the Dominant topic, Perc Contribution and Keywords for each document
            topic_percs = 0
            for j, (topic_num, prop_topic) in enumerate(row):
                if j == 0:  # => dominant topic
                    topic_percs = prop_topic
                    dominant_topic = topic_percs
                    k=0
                    for word, prob in mdl.get_topic_words(topic_num):
                        if (k < 3):
                            topic_top3words[topic_num].append(word)
                        k+=1
                else:
                    break

            dominant_topics.append((i, dominant_topic))
            topic_percentages.append(topic_percs)

        # Distribution of Dominant Topics in Each Document
        df = pd.DataFrame(dominant_topics, columns=['Document_Id', 'Dominant_Topic'])
        dominant_topic_in_each_doc = df.groupby('Dominant_Topic').size()
        df_dominant_topic_in_each_doc = dominant_topic_in_each_doc.to_frame(name='count').reset_index()

        # Total Topic Distribution by actual weight
        topic_weightage_by_doc = pd.DataFrame([dict(t) for t in topic_percentages])
        df_topic_weightage_by_doc = topic_weightage_by_doc.sum().to_frame(name='count').reset_index()

        df_top3words_stacked = pd.DataFrame(topic_top3words, columns=['topic_id', 'words'])
        df_top3words = df_top3words_stacked.groupby('topic_id').agg(', \n'.join)
        df_top3words.reset_index(level=0, inplace=True)

        print(df_top3words.head(5))

        # Plot
        fig, (ax1, ax2) = plt.subplots(1, 2, figsize=(10, 4), dpi=120, sharey=True)

        # Topic Distribution by Dominant Topics
        ax1.bar(x='Dominant_Topic', height='count', data=df_dominant_topic_in_each_doc, width=.5, color='firebrick')
        ax1.set_xticks(range(df_dominant_topic_in_each_doc.Dominant_Topic.unique().__len__()))
        tick_formatter = FuncFormatter(lambda x, pos: 'Topic ' + str(x)+ '\n' + df_top3words.loc[df_top3words.topic_id==x, 'words'].values[0])
        ax1.xaxis.set_major_formatter(tick_formatter)
        ax1.set_title('Number of Documents by Dominant Topic', fontdict=dict(size=10))
        ax1.set_ylabel('Number of Documents')
        ax1.set_ylim(0, 1000)

        # Topic Distribution by Topic Weights
        ax2.bar(x='index', height='count', data=df_topic_weightage_by_doc, width=.5, color='steelblue')
        ax2.set_xticks(range(df_topic_weightage_by_doc.index.unique().__len__()))
        ax2.xaxis.set_major_formatter(tick_formatter)
        ax2.set_title('Number of Documents by Topic Weightage', fontdict=dict(size=10))

        plt.show()

    def tSNE(self, mdl, matrix, label, topic_number=10):
        from bokeh.plotting import figure, output_file, show
        from bokeh.models import Label
        from bokeh.io import output_notebook
        import matplotlib.colors as mcolors
        from sklearn.manifold import TSNE

        # Array of topic weights
        arr = pd.DataFrame(matrix).fillna(0).values

        # Dominant topic number in each doc
        topic_num = np.argmax(arr, axis=1)

        # tSNE Dimension Reduction
        tsne_model = TSNE(n_components=2, verbose=1, random_state=0, angle=.99, init='pca')
        tsne_lda = tsne_model.fit_transform(arr)

        n_topics = topic_number
        mycolors = np.array([color for name, color in matplotlib.colors.cnames.items()])
        plot = figure(title="t-SNE Clustering of {} " + label + "Topics".format(n_topics),
                      plot_width=900, plot_height=700)

        plot.scatter(x=tsne_lda[:,0], y=tsne_lda[:,1], color=mycolors[topic_num])

        show(plot)

    def pivot_and_smooth(self, df, rows_variable, smooth_value, cols_variable, values_variable):
        """
        Turns the pandas dataframe into a data matrix.
        Args:
            df (dataframe): aggregated dataframe
            smooth_value (float): value to add to the matrix to account for the priors
            rows_variable (str): name of dataframe column to use as the rows in the matrix
            cols_variable (str): name of dataframe column to use as the columns in the matrix
            values_variable(str): name of the dataframe column to use as the values in the matrix
        Returns:
            dataframe: pandas matrix that has been normalized on the rows.
        """
        matrix = df.pivot(index=rows_variable, columns=cols_variable, values=values_variable).fillna(value=0)
        matrix = matrix.values + smooth_value

        normed = sklearn.preprocessing.normalize(matrix, norm='l1', axis=1)

        return pd.DataFrame(normed)

    def make_pyLDAVis(self, mdl, matrix, documents):
        from collections import Counter
        data_flat = [w for w_list in matrix for w in w_list]
        counter = Counter(data_flat)

        out = []
        for k in range(mdl.k):
            for word, weight in mdl.get_topic_words(k):
                #print(word + " : " + str(weight))
                out.append([k, word, weight])

        df = pd.DataFrame(out, columns=['topic_id', 'word', 'importance'])

        smooth_value = mdl.eta

        phi_df = self.pivot_and_smooth(df, 'topic_id', smooth_value, 'word', 'importance')
        #phi_df = phi_df.sort_values(by='word', ascending=True)
        print(phi_df[:10])

        # Get vocab and term frequencies from statefile
        vocab = df['word'].value_counts().reset_index()
        vocab.columns = ['word', 'word_count']
        vocab = vocab.sort_values(by='word', ascending=True)
        print(vocab[:10])

        out = []
        doc_id = 0
        for d in mdl.docs:
            out.append([doc_id, len(documents[doc_id])])
            doc_id += 1
        docs = pd.DataFrame(out, columns=['#doc', 'doc_length'])
        #print(docs[:10])
        print(str(len(docs)) + " :: " + str(len(docs.columns)) )
        mat = []
        docs_topics_df = pd.DataFrame()
        for d in mdl.docs:
            topic_dist = d.get_topic_dist()
            # Get the Dominant topic, Perc Contribution and Keywords for each document
            row = []
            for prop_topic in topic_dist:
                row.append(prop_topic)
            mat.append(row)

        matrix = np.transpose(matrix)
        Dict = {}
        topic_id = 0
        for r in matrix:
            j = 0
            Dict[str(topic_id)] = {}
            for c in r:
                # Adding elements one at a time
                Dict[str(topic_id)][str(j)] = c
                j += 1
            topic_id += 1
        docs_topics_df = pd.DataFrame(Dict)

        #print(docs_topics_df[:10])
        #print(str(len(docs_topics_df)) + " :: " + str(len(docs_topics_df.columns)))

        data = {'topic_term_dists': phi_df,
                'doc_topic_dists': docs_topics_df,
                'doc_lengths': list(docs['doc_length']),
                'vocab': list(vocab['word']),
                'term_frequency': list(vocab['word_count'])
                }

        import pyLDAvis
        vis_data = pyLDAvis.prepare(**data)
        #pyLDAvis.display(vis_data)

        pyLDAvis.save_html(vis_data, 'vis.html')

    def hdp_model(self, text_data, save_path):
        mdl = tp.HDPModel(tw=tp.TermWeight.ONE, min_cf=3, rm_top=5)
        index = 0
        for doc in text_data:
            print(str(index) + " : " + str(doc))
            mdl.add_doc(doc)
            index += 1

        mdl.burn_in = 100
        mdl.train(0)
        print('Num docs:', len(mdl.docs), ', Vocab size:', mdl.num_vocabs, ', Num words:', mdl.num_words)
        print('Removed top words:', mdl.removed_top_words)
        print('Training...', file=sys.stderr, flush=True)
        for i in range(0, 1000, 10):
            mdl.train(10)
            print('Iteration: {}\tLog-likelihood: {}\tNum. of topics: {}'.format(i, mdl.ll_per_word, mdl.live_k))

        print('Saving...', file=sys.stderr, flush=True)
        mdl.save(save_path, True)

        topic_num = 0
        # extract candidates for auto topic labeling
        extractor = tp.label.PMIExtractor(min_cf=10, min_df=5, max_len=5, max_cand=10000)
        cands = extractor.extract(mdl)

        # ranking the candidates of labels for a specific topic
        labeler = tp.label.FoRelevance(mdl, cands, min_df=5, smoothing=1e-2, mu=0.25)
        important_topics = [k for k, v in
                            sorted(enumerate(mdl.get_count_by_topics()), key=lambda x: x[1], reverse=True)]
        for k in important_topics:
            if not mdl.is_live_topic(k): continue
            print("== Topic #{} ==".format(k))
            print("Labels:", ', '.join(label for label, score in labeler.get_topic_labels(k, top_n=5)))
            for word, prob in mdl.get_topic_words(k, top_n=10):
                print(word, prob, sep='\t')
            print()
            topic_num+=1
        return (mdl, topic_num)

    def hlda_model(self, text_data, save_path):
        mdl = tp.HLDAModel(tw=tp.TermWeight.ONE, min_cf=3, rm_top=5, depth=2)
        index = 0
        for doc in text_data:
            print(str(index) + " : " + str(doc))
            mdl.add_doc(doc)
            index += 1

        mdl.burn_in = 100
        mdl.train(0)
        print('Num docs:', len(mdl.docs), ', Vocab size:', mdl.num_vocabs, ', Num words:', mdl.num_words)
        print('Removed top words:', mdl.removed_top_words)
        print('Training...', file=sys.stderr, flush=True)
        for i in range(0, 1000, 10):
            mdl.train(10)
            print('Iteration: {}\tLog-likelihood: {}\tNum. of topics: {}'.format(i, mdl.ll_per_word, mdl.live_k))

        print('Saving...', file=sys.stderr, flush=True)
        mdl.save(save_path, True)

        # extract candidates for auto topic labeling
        extractor = tp.label.PMIExtractor(min_cf=10, min_df=5, max_len=5, max_cand=10000)
        cands = extractor.extract(mdl)

        # ranking the candidates of labels for a specific topic
        labeler = tp.label.FoRelevance(mdl, cands, min_df=5, smoothing=1e-2, mu=0.25)
        important_topics = [k for k, v in
                            sorted(enumerate(mdl.get_count_by_topics()), key=lambda x: x[1], reverse=True)]
        for k in important_topics:
            if not mdl.is_live_topic(k): continue
            print("== Topic #{} ==".format(k))

            children_ids = mdl.children_topics(k)
            for id in children_ids:
                print('children topic ' + str(id) + ' ')
            print('\n')

            print("Labels:", ', '.join(label for label, score in labeler.get_topic_labels(k, top_n=5)))
            for word, prob in mdl.get_topic_words(k, top_n=10):
                print(word, prob, sep='\t')
            print()
        return mdl

    def lda_model(self, text_data, save_path, topic_number=20):
        mdl = tp.LDAModel(tw=tp.TermWeight.ONE, min_cf=3, rm_top=5, k=topic_number)
        index=0
        for doc in text_data:
            print(str(index) + " : " + str(doc))
            mdl.add_doc(doc)
            index+=1


        mdl.burn_in = 100
        mdl.train(0)
        print('Num docs:', len(mdl.docs), ', Vocab size:', mdl.num_vocabs, ', Num words:', mdl.num_words)
        print('Removed top words:', mdl.removed_top_words)
        print('Training...', file=sys.stderr, flush=True)
        for i in range(0, 1500, 10):
            mdl.train(10)
            print('Iteration: {}\tLog-likelihood: {}'.format(i, mdl.ll_per_word))

        print('Saving...', file=sys.stderr, flush=True)
        mdl.save(save_path, True)

        # extract candidates for auto topic labeling
        extractor = tp.label.PMIExtractor(min_cf=10, min_df=5, max_len=5, max_cand=10000)
        cands = extractor.extract(mdl)

        # ranking the candidates of labels for a specific topic
        labeler = tp.label.FoRelevance(mdl, cands, min_df=5, smoothing=1e-2, mu=0.25)
        for k in range(mdl.k):
            print("== Topic #{} ==".format(k))
            print("Labels:", ', '.join(label for label, score in labeler.get_topic_labels(k, top_n=5)))
            for word, prob in mdl.get_topic_words(k, top_n=10):
                print(word, prob, sep='\t')
            print()

        return mdl

    def dmr_model(self, text_data, pair_map, save_path, topic_number=20):
        mdl = tp.DMRModel(tw=tp.TermWeight.ONE, min_cf=3, rm_top=5, k=topic_number)
        print(mdl.perplexity)

        index=0
        for doc in text_data:
            print(str(index) + " : " + str(doc))
            year=pair_map[index]
            mdl.add_doc(doc,metadata=year)
            index+=1

        mdl.burn_in = 100
        mdl.train(0)
        print('Num docs:', len(mdl.docs), ', Vocab size:', mdl.num_vocabs, ', Num words:', mdl.num_words)
        print('Removed top words:', mdl.removed_top_words)
        print('Training...', file=sys.stderr, flush=True)
        for i in range(0, 1000, 10):
            mdl.train(10)
            print('Iteration: {}\tLog-likelihood: {}'.format(i, mdl.ll_per_word))

        print('Saving...', file=sys.stderr, flush=True)
        mdl.save(save_path, True)

        # extract candidates for auto topic labeling
        extractor = tp.label.PMIExtractor(min_cf=10, min_df=5, max_len=5, max_cand=10000)
        cands = extractor.extract(mdl)

        # ranking the candidates of labels for a specific topic
        labeler = tp.label.FoRelevance(mdl, cands, min_df=5, smoothing=1e-2, mu=0.25)
        for k in range(mdl.k):
            print("== Topic #{} ==".format(k))
            print("Labels:", ', '.join(label for label, score in labeler.get_topic_labels(k, top_n=5)))
            for word, prob in mdl.get_topic_words(k, top_n=10):
                print(word, prob, sep='\t')
            print()

        # Init output
        topics_features = pd.DataFrame()
        col_features = []

        for k in range(mdl.k):
            print('Topic #{}'.format(k))
            array_features=[]
            features={}
            for m in range(mdl.f):
                #print('feature ' + mdl.metadata_dict[m] + " --> " + str(mdl.lambdas[k][m]) + " ")
                features[mdl.metadata_dict[m]]=mdl.lambdas[k][m]
                array_features.append(mdl.lambdas[k][m])
                if int(k) == 0:
                    #print('feature ' + mdl.metadata_dict[m])
                    col_features.append(mdl.metadata_dict[m])

            a = np.array(array_features)
            median=np.median(a)
            max=np.max(a)
            min=np.min(a)

            new_features=[]
            #new_features.append(k)
            for col in col_features:
                val = features[col]
                final_val = abs(max) + val + abs(median)
                features[col] = final_val
                #print("YYYYYY " + col + " :: " + str(features[col]))
                new_features.append(final_val)

            topics_features = topics_features.append(pd.Series(new_features), ignore_index=True)
            print("median " + str(median) + " : " + str(max) + " : " + str(min))

            for word, prob in mdl.get_topic_words(k):
                print('\t', word, prob, sep='\t')

        col_feaures = sorted(col_features, reverse=False)
        #col_features.insert(0,'Topic ID')
        topics_features.columns = col_features

        topics_features.to_csv('dmr_topic_year.csv', sep=',', encoding='utf-8')
        print(topics_features.head(20))

        df1_transposed = topics_features.T.rename_axis('Date').reset_index()
        #labels = []
        #for i in range(0,topic_number-1):
        #    labels.append('Topic_'+str(i))
        #df1_transposed.columns=labels

        import seaborn as sns
        import matplotlib.colors as mcolors

        print(df1_transposed.head(20))
        df1_transposed = df1_transposed.melt('Date', var_name='Topic', value_name='Importance Score')
        g = sns.relplot(x="Date", y="Importance Score", hue='Topic', dashes=False, markers=True,  data=df1_transposed, kind='line')

        output='dmr_topic.png'
        g.fig.suptitle('DMR Topic Model Results')
        g.savefig(output, format='png', dpi=500)
        # Show the plot
        plt.show()

        return mdl

    def inferLDATopicModel(self, model_file, unseen_words):

        mdl = tp.LDAModel.load(model_file)

        doc_inst = mdl.make_doc(unseen_words)
        topic_dist, ll = mdl.infer(doc_inst)
        print("Topic Distribution for Unseen Docs: ", topic_dist)
        print("Log-likelihood of inference: ", ll)

if __name__ == '__main__':

    import pyTextMiner as ptm

    mecab_path='C:\\mecab\\mecab-ko-dic'
    pipeline = ptm.Pipeline(ptm.splitter.NLTK(),
                            ptm.tokenizer.MeCab(mecab_path),
                            ptm.helper.POSFilter('NN*'),
                            ptm.helper.SelectWordOnly(),
                            ptm.helper.StopwordFilter(file='../stopwords/stopwordsKor.txt'))

    corpus = ptm.CorpusFromFieldDelimitedFileWithYear('../mallet/topic_input/sample_dmr_input.txt',doc_index=2,year_index=1)
    pair_map = corpus.pair_map

    result = pipeline.processCorpus(corpus.docs)
    text_data = []
    for doc in result:
        new_doc = []
        for sent in doc:
            for _str in sent:
                if len(_str) > 0:
                    new_doc.append(_str)
        text_data.append(new_doc)

    topic_number=20
    print('Running DMR')
    dmr_model_name='../test.lda.bin'
    topic_model = pyTextMinerTopicModel()

    #topic_model.dmr_model(text_data, pair_map, dmr_model_name, topic_number)
    topic_model.hdp_model(text_data, dmr_model_name)

    unseen_text='아사이 베리 블루베리 비슷하다'
    #topic_model.inferLDATopicModel(dmr_model_name, unseen_text)



class POSFilter:
    IN_TYPE = [list, tuple]
    OUT_TYPE = [list, tuple]

    def __init__(self, *posWanted):
        import re
        self.wanted = set(p for p in posWanted if not p.endswith('*'))
        self.re = re.compile('(' + '|'.join(p[:-1] for p in posWanted if p.endswith('*')) + ').*')

    def test(self, pos):
        if pos in self.wanted: return True
        if self.re.match(pos): return True
        return False

    def __call__(self, *args, **kwargs):
        return [i for i in args[0] if self.test(i[1])]

class StopwordFilter:
    IN_TYPE = [list, str]
    OUT_TYPE = [list, str]

    def __init__(self, stopwords = [], file = None):
        if file:
            stopwords = stopwords + [line.strip() for line in open(file, encoding='utf-8')]
        self.stopwords = set(stopwords)
        self.stopwordsPrefix = ('http', 'https', 'ftp', 'git', 'thatt')

    def __call__(self, *args, **kwargs):
        #any(e for e in test_list if e.startswith('three') or e.endswith('four'))
        return [i for i in args[0] if i.lower() not in self.stopwords and (i.lower().startswith(tuple(p for p in self.stopwordsPrefix)) == False)]

class SelectWordOnly:
    IN_TYPE = [tuple]
    OUT_TYPE = [str]

    def __call__(self, *args, **kwargs):
        return args[0][0]

class ToLowerCase:
    IN_TYPE = [str]
    OUT_TYPE = [str]

    def __call__(self, *args, **kwargs):
        return args[0].lower()


# -*- coding: UTF-8 -*-

import io
import re


class DictionarySentimentAnalyzer:
    def __init__(self):
        name = 'DictionarySentimentAnalyzer'
        self.dict_list={}

    def readPolarityDictionary(self, file):
        fh = open(file, mode="r", encoding="utf-8")
        fh.readline()
        for line in fh:
            fields = line.split(',')
            n_token = ''
            toks = fields[0].split(";")
            for tok in toks:
                n_token += tok.split("/")[0]
            print("element " + n_token + " -- " + fields[len(fields)-2] + " : " + fields[len(fields)-1])
            sent_dict = {}
            sent_dict['word'] = n_token
            sent_dict['polarity'] = fields[len(fields)-2]
            if sent_dict['polarity'] is 'NEG':
                sent_dict['score'] = -float(fields[len(fields)-1].replace("\n",""))
            elif sent_dict['polarity'] is 'POS':
                sent_dict['score'] = float(fields[len(fields) - 1].replace("\n", ""))
            else:
                sent_dict['score'] = 0.0

            if n_token not in self.dict_list:
                self.dict_list[n_token] = sent_dict

    def readKunsanUDictionary(self, file):
        with open(file, encoding='utf-8-sig', mode='r') as f:
            for line in f:
                fields = line.split('\t')
                escaped = fields[0]
                sent_dict = {}
                if len(fields) < 2: continue

                if escaped not in self.dict_list:
                    sent_dict['word'] = escaped
                    sent_dict['score'] = float(fields[1])
                    print('term ' + escaped + " : " + fields[1])
                    if float(fields[1]) < 0:
                        sent_dict['polarity'] = 'NEG'
                    elif float(fields[1]) > 0:
                        sent_dict['polarity'] = 'POS'
                    else:
                        sent_dict['polarity'] = 'NEU'

                    self.dict_list[escaped] = sent_dict

    def readCurseDictionary(self, file):
        with open(file, encoding='utf-8-sig', mode='r') as f:
            for line in f:
                term = line.strip()
                sent_dict={}
                if term not in self.dict_list:
                    sent_dict['word'] = term
                    sent_dict['score'] = -2.0
                    print('term ' + term)
                    sent_dict['polarity'] = 'NEG'

                    self.dict_list[term] = sent_dict

    def readNegativeDictionary(self, file):
        with open(file, encoding='utf-8-sig', mode='r') as f:
            for line in f:
                term = line.strip()
                sent_dict={}
                if term not in self.dict_list:
                    sent_dict['word'] = term
                    sent_dict['score'] = -1.0
                    print('term ' + term)
                    sent_dict['polarity'] = 'NEG'

                    self.dict_list[term] = sent_dict

    def readPositiveDictionary(self, file):
        with open(file, encoding='utf-8-sig', mode='r') as f:
            for line in f:
                term = line.strip()
                sent_dict={}
                if term not in self.dict_list:
                    sent_dict['word'] = term
                    sent_dict['score'] = 1.0
                    print('term ' + term)
                    sent_dict['polarity'] = 'POS'
                    self.dict_list[term] = sent_dict

    def readPositiveEmotiDictionary(self, file):
        with open(file, encoding='utf-8-sig', mode='r') as f:
            for line in f:
                term = line.strip()
                sent_dict={}
                if term not in self.dict_list:
                    sent_dict['word'] = term
                    sent_dict['score'] = 1.0
                    print('term ' + term)
                    sent_dict['polarity'] = 'POS'
                    self.dict_list[term] = sent_dict

    def readNegativeEmotiDictionary(self, file):
        with open(file, encoding='utf-8-sig', mode='r') as f:
            for line in f:
                term = line.strip()
                sent_dict={}
                if term not in self.dict_list:
                    sent_dict['word'] = term
                    sent_dict['score'] = -1.0
                    print('term ' + term)
                    sent_dict['polarity'] = 'NEG'
                    self.dict_list[term] = sent_dict

    def getSentiDictionary(self):
        return self.dict_list

if __name__ == '__main__':
    import pyTextMiner as ptm
    import io
    import nltk

    sentiAnalyzer = DictionarySentimentAnalyzer()

    file_name = './data/SentiWord_Dict.txt'
    sentiAnalyzer.readKunsanUDictionary(file_name)
    file_name = './data/korean_curse_words.txt'
    sentiAnalyzer.readCurseDictionary(file_name)
    file_name = './data/negative_words_ko.txt'
    sentiAnalyzer.readNegativeDictionary(file_name)
    file_name = './data/positive_words_ko.txt'
    sentiAnalyzer.readPositiveDictionary(file_name)
    file_name = './data/emo_negative.txt'
    sentiAnalyzer.readNegativeEmotiDictionary(file_name)
    file_name = './data/emo_positive.txt'
    sentiAnalyzer.readPositiveEmotiDictionary(file_name)
    file_name = './data/polarity.csv'
    sentiAnalyzer.readPolarityDictionary(file_name)

    dict_list = sentiAnalyzer.getSentiDictionary()

    pipeline = None
    corpus = ptm.CorpusFromFieldDelimitedFile('../data/donald.txt',2)
    mecab_path = 'C:\\mecab\\mecab-ko-dic'
    mode = 'korean_lemmatizer'
    if mode is not 'korean_lemmatizer':
         pipeline = ptm.Pipeline(ptm.splitter.NLTK(),
                             ptm.tokenizer.MeCab(mecab_path),
                             #ptm.tokenizer.Komoran(),
                             ptm.helper.SelectWordOnly(),
                             #ptm.ngram.NGramTokenizer(1,2,concat=' '),
                             ptm.helper.StopwordFilter(file='../stopwords/stopwordsKor.txt'))
    else :
         pipeline = ptm.Pipeline(ptm.splitter.NLTK(),
                             ptm.tokenizer.MeCab(mecab_path),
                             #ptm.tokenizer.Komoran(),
                             ptm.lemmatizer.SejongPOSLemmatizer(),
                             ptm.helper.SelectWordOnly(),
                             #ptm.ngram.NGramTokenizer(1, 2, concat=' '),
                             ptm.helper.StopwordFilter(file='../stopwords/stopwordsKor.txt'))

    documents = ['오늘은 비가와서 그런지 매우 우울하다',
                 'KBO 일명 잘나가는 야구선수들 대부분 팬서비스는 뒷전... 니들 연봉은 팬들이 주는거다!!',
                 '시험이 끝나야 놀지 스트레스 받아ㅠㅠ',
                 '꽤 잘 만들어진 좀비 영화라는 생각이 듬',
                 '옷을 잔뜩 사서 완전 무거워서 신나',
                 '행복한 하루의 끝이라 않좋네!',
                 '더운날에는 아이스커피가 무섭지 않다~~']

    #result = pipeline.processCorpus(corpus)
    result = pipeline.processCorpus(documents)
    print(len(corpus.docs))

    for doc in result:
        total_score = 0.0
        count = 0

        for sent in doc:
            idx = 0
            for _str in sent:
                if len(_str) > 0:
                    score = 0.0
                    flag = False
                    if len(_str.split(' ')) > 0:
                        if _str.split(' ')[0] == '안':
                            _str = _str.split(' ')[1]
                            flag = True
                    dictionary_ele = dict_list.get(_str)
                    if (dictionary_ele != None):
                        polarity = dictionary_ele.get('polarity')
                        if flag:
                            score = -float(dictionary_ele.get('score'))
                        else:
                            score = float(dictionary_ele.get('score'))

                        count += 1
                        #print(_str + " == " + polarity + " " + str(score))
                        total_score += score
                    else:
                        total_score += score

                idx+=1

        if (count != 0):
            avg_score = total_score/count
            print("AVG SCORE " + str(avg_score) + " for " + str(doc))


"""
original: https://github.com/sublee/korean/blob/master/korean/hangul.py (sub@subl.ee)
modified: https://github.com/bage79/nlp4kor/blob/master/bage_utils/hangul_util.py (bage79@gmail.com)
"""
import random
import re
import traceback
import warnings

import numpy

from treform.utility.string_util import StringUtil


def to_sequence(*sequences):
    def to_tuple(sequence):
        if not sequence:
            return sequence,
        return tuple(sequence)

    return sum(map(to_tuple, sequences), ())


def _to_one_hot_vector(c: str, li=None) -> numpy.ndarray:
    """
    :param c: 문자
    :param li: 문자를 찾을 리스트 e.g. ['ㄱ', 'ㄴ', 'ㄷ', ...]
    :return: one hot vector (numpy.ndarray)
    """
    if li is None:
        li = []

    try:
        idx = li.index(c)
    except:
        raise Exception('%s not found in list' % c)

    a = numpy.zeros(len(li), dtype=numpy.int32)
    a[idx] = 1
    return a


warnings.filterwarnings("ignore", category=FutureWarning, append=1)  # for warning in re
han_eng_re = re.compile('[^ㄱ-ㅎㅏ-ㅣ가-힣0-9a-zA-Z\\s]')
han_eng_symbol_re = re.compile('[^ㄱ-ㅎㅏ-ㅣ가-힣0-9a-zA-Z~!@#=^$%&_+:\";\',.?/(){\}\[\]\\s]')


# noinspection PyGlobalUndefined
class HangulUtil(object):
    """
    original: https://github.com/sublee/korean/blob/master/korean/hangul.py
    modified: bage79@gmail.com
    """

    IS_NOT_HANGUL = -1
    global to_sequence

    JA_LIST = to_sequence(u'ㄱㄲㄳㄴㄵㄶㄷㄸㄹㄺㄻㄼㄽㄾㄿㅀㅁㅂㅃㅄㅅㅆㅇㅈㅉㅊㅋㅌㅍㅎ')  # 자음
    MO_LIST = to_sequence(u'ㅏㅐㅑㅒㅓㅔㅕㅖㅗㅘㅙㅚㅛㅜㅝㅞㅟㅠㅡㅢㅣ')  # 모음
    WANSUNG_LIST = tuple([chr(c) for c in range(ord(u'가'), ord(u'힣') + 1)])  # 완성형

    JA_RANGE = [ord(c) for c in JA_LIST]  # 자음 범위
    MO_RANGE = [ord(c) for c in MO_LIST]  # 모음 범위
    WANSUNG_RANGE = range(ord(u'가'), ord(u'힣') + 1)  # 완성형

    HANGUL_LIST = list(set(list(MO_LIST) + list(JA_LIST) + [chr(c) for c in WANSUNG_RANGE]))
    HANGUL_LIST.sort()
    HANGUL_FIRST = WANSUNG_RANGE[0]

    CHAR2ONE_HOT = list()

    CHO_LIST = to_sequence(u'ㄱㄲㄴㄷㄸㄹㅁㅂㅃㅅㅆㅇㅈㅉㅊㅋㅌㅍㅎ')  # 초성
    JUNG_LIST = MO_LIST  # 중성
    JONG_LIST = to_sequence(u'', u'ㄱㄲㄳㄴㄵㄶㄷㄹㄺㄻㄼㄽㄾㄿㅀㅁㅂㅄㅅㅆㅇㅈㅊㅋㅌㅍㅎ')  # 종성

    CHO_NOISE_LIST = to_sequence(u'ㄱㄴㄷㄹㅁㅂㅅㅇㅈㅊㅋㅌㅍㅎ')  # 초성
    JUNG_NOISE_LIST = to_sequence(u'ㅏㅑㅓㅕㅗㅛㅜㅠㅡㅣ')  # 모음
    JONG_NOISE_LIST = to_sequence(u'', u'ㄱㄴㄷㄹㅁㅂㅅㅇㅈㅊㅋㅌㅍㅎ')  # 종성

    CHO_JUNG_JONG_LIST = (CHO_LIST, JUNG_LIST, JONG_LIST)  # 초성 중성 종성
    CHO_RANGE = [ord(c) for c in CHO_LIST if len(c) > 0]  # 초성 범위
    JUNG_RANGE = [ord(c) for c in JUNG_LIST if len(c) > 0]  # 중성 범위
    JONG_RANGE = [ord(c) for c in JONG_LIST if len(c) > 0]  # 종성 범위

    HANJA_RANGE = range(ord(u'一'), ord(u'龥') + 1)
    HANJA_LIST = [c for c in HANJA_RANGE]

    ENGLISH_LOWER_RANGE = range(ord(u'a'), ord(u'z') + 1)
    ENGLISH_UPPER_RANGE = range(ord(u'A'), ord(u'Z') + 1)
    ENGLISH_LIST = [chr(c) for c in ENGLISH_LOWER_RANGE] + [chr(c) for c in ENGLISH_UPPER_RANGE]

    NUM_LIST = [str(c) for c in range(10)]
    KEYBOARD_SPECIAL_LIST = list(to_sequence(r"""`~!@#$%^&*()_+-+[]\;',./{}|:"<>?"""))

    CHAR_LIST = HANGUL_LIST + ENGLISH_LIST + NUM_LIST + KEYBOARD_SPECIAL_LIST

    KEYBOARD_ENG_LIST = 'qwertyuiopasdfghjklzxcvbnm' + 'QWERTOP'
    KEYBOARD_HAN_LIST = 'ㅂㅈㄷㄱㅅㅛㅕㅑㅐㅔㅁㄴㅇㄹㅎㅗㅓㅏㅣㅋㅌㅊㅍㅠㅜㅡ' + 'ㅃㅉㄸㄲㅆㅒㅖ'
    KEYBOARD_ENG_TO_HAN = dict(zip(KEYBOARD_ENG_LIST, KEYBOARD_HAN_LIST))

    del to_sequence

    @classmethod
    def to_cho_jung_jong_vector(cls, c: str):
        """
        초성(19) + 중성(21) + 종성(28) + 영어(52) + 숫자(10) + 특수문자32)
        :param c:
        :return:
        """
        hangul_vector = numpy.zeros(len(cls.CHO_LIST) + len(cls.JUNG_LIST) + len(cls.JONG_LIST))
        english_vector = numpy.zeros(len(cls.ENGLISH_LIST))
        num_vector = numpy.zeros(len(cls.NUM_LIST))
        special_vector = numpy.zeros(len(cls.KEYBOARD_SPECIAL_LIST))

        if cls.is_hangul_char(c):
            cho, jung, jong = cls.split2cho_jung_jong(c)
            if len(cho) == 1:
                cho_vector = _to_one_hot_vector(cho, cls.CHO_LIST)
            else:
                cho_vector = numpy.zeros(len(cls.CHO_LIST))
            if len(jung) == 1:
                jung_vector = _to_one_hot_vector(jung, cls.JUNG_LIST)
            else:
                jung_vector = numpy.zeros(len(cls.JUNG_LIST))
            if len(jong) == 1:
                jong_vector = _to_one_hot_vector(jong, cls.JONG_LIST)
            else:
                jong_vector = numpy.zeros(len(cls.JONG_LIST))
            hangul_vector = numpy.concatenate((cho_vector, jung_vector, jong_vector))
        elif c in cls.ENGLISH_LIST:
            english_vector = _to_one_hot_vector(c, cls.ENGLISH_LIST)
        elif c in cls.NUM_LIST:
            num_vector = _to_one_hot_vector(c, cls.NUM_LIST)
        elif c in cls.KEYBOARD_SPECIAL_LIST:
            special_vector = _to_one_hot_vector(c, cls.KEYBOARD_SPECIAL_LIST)

        return numpy.concatenate((hangul_vector, english_vector, num_vector, special_vector))

    # noinspection PyUnresolvedReferences
    @classmethod
    def to_one_hot_vector(cls, c: str):
        return cls.CHAR2ONE_HOT.get(c, None)

    @classmethod
    def to_one_hot_index(cls, c: str):
        try:
            return cls.CHAR_LIST.index(c)
        except:  # not found
            return -1

    @classmethod
    def load(cls):
        cls.CHAR2ONE_HOT = dict([(c, _to_one_hot_vector(c, li=cls.CHAR_LIST)) for c in cls.CHAR_LIST])
        # print('keys:', cls.HANGUL_ONE_HOT.keys())

    @classmethod
    def to_cho_index(cls, c: str):
        try:
            return cls.CHO_LIST.index(c)
        except:  # not found
            return -1

    @classmethod
    def to_jung_index(cls, c: str):
        try:
            return cls.JUNG_LIST.index(c)
        except:  # not found
            return -1

    @classmethod
    def to_jong_index(cls, c: str):
        try:
            return cls.JONG_LIST.index(c)
        except:  # not found
            return -1

    @staticmethod
    def remain_han_eng_symbol(s):
        """
        한글, 영어, 숫자, 일부 기호만 남기고 제거
        :param s:
        :return:
        """
        return ' '.join(han_eng_symbol_re.sub('', s).split(' ')).strip()

    @staticmethod
    def remain_han_eng(s):
        """
        한글, 영어, 숫자, 일부 기호만 남기고 제거
        :param s:
        :return:
        """
        return ' '.join(han_eng_re.sub('', s).split(' ')).strip()

    @classmethod
    def has_hangul(cls, word: list) -> object:
        """
        :param word: 단어
        :return: 한글 음절을 포함하는지 여부
        """
        try:
            for char in word:
                if (ord(char) in cls.WANSUNG_RANGE) or (ord(char) in cls.JA_RANGE) or (ord(char) in cls.MO_RANGE):
                    return True
            return False
        except:
            print(traceback.format_exc())

    @classmethod
    def has_english(cls, word):
        """
        :param word: 단어
        :return: 영어 음절을 포함하는지 여부
        """
        try:
            for char in word:
                if ord(char) in cls.ENGLISH_LOWER_RANGE or ord(char) in cls.ENGLISH_UPPER_RANGE:
                    return True
            return False
        except:
            print(traceback.format_exc())

    @classmethod
    def __char_offset(cls, char):
        """
        :param char: 음절
        :return: offset from u"가".
        """
        if isinstance(char, int):
            offset = char
        else:
            if len(char) != 1:
                return cls.IS_NOT_HANGUL
            elif not cls.is_full_hangul(char):
                return cls.IS_NOT_HANGUL
            offset = ord(char) - cls.HANGUL_FIRST

        if offset >= len(cls.WANSUNG_RANGE):
            return cls.IS_NOT_HANGUL
        else:
            return offset

    @classmethod
    def is_full_hangul(cls, word, exclude_chars='.,'):
        """
        :param exclude_chars: 예외로 둘 문자들 (공백은 기본 포함)
        :param word: 단어
        :return: 모든 음절이 한글인지 여부
        """
        try:
            for char in word:
                if char != ' ' and (char not in exclude_chars) and (ord(char) not in cls.WANSUNG_RANGE):
                    return False
            return True
        except:
            print(traceback.format_exc())

    @classmethod
    def is_full_hangul_or_english(cls, word, exclude_chars=''):
        """
        :param exclude_chars: 예외로 둘 문자들 (공백은 기본 포함)
        :param word: 단어
        :return: 모든 음절이 한글 또는 영어인지 여부
        """
        # print('is_full_hangul(%s)' % word)
        # print('is_full_hangul;', id(log), log.level)
        try:
            for char in word:
                if char != ' ' and (char not in exclude_chars) and (ord(char) not in cls.WANSUNG_RANGE) and (ord(char) not in cls.ENGLISH_UPPER_RANGE) and (
                        ord(char) not in cls.ENGLISH_LOWER_RANGE):
                    return False
            return True
        except:
            print(traceback.format_exc())

    @classmethod
    def has_hangul(cls, word):
        """
        :param word: 단어
        :return: 한글 음절을 포함하는지 여부
        """
        try:
            for char in word:
                if ord(char) in cls.WANSUNG_RANGE:
                    return True
            return False
        except:
            print(traceback.format_exc())

    @classmethod
    def has_hanja(cls, word):
        """
        :param word: 단어
        :return: 한자 음절을 포함하는지 여부
        """
        try:
            for char in word:
                if ord(char) in cls.HANJA_RANGE:
                    return True
            return False
        except:
            print(traceback.format_exc())

    @classmethod
    def is_hangul_char(cls, char):
        """

        :param char: 음절
        :return: 한글 여부
        """
        if len(char) == 0 or len(char) > 1:
            return False
        try:
            if char in cls.HANGUL_LIST:  # if ord(char) in cls.WANSUNG_RANGE:
                return True
            return False
        except:
            print(traceback.format_exc())

    @classmethod
    def is_english_char(cls, char):
        """

        :param char: 음절
        :return: 영어 여부
        """
        if len(char) == 0 or len(char) > 1:
            return False
        try:
            if ord(char) in cls.ENGLISH_LOWER_RANGE or ord(char) in cls.ENGLISH_UPPER_RANGE:
                return True
            return False
        except:
            print(traceback.format_exc())

    @classmethod
    def is_hanja_char(cls, char):
        """

        :param char: 음절
        :return: 한자 여부
        """
        if len(char) == 0 or len(char) > 1:
            return False
        try:
            if ord(char) in cls.HANJA_RANGE:
                return True
            return False
        except:
            print(traceback.format_exc())

    @classmethod
    def is_moum(cls, char):
        """
        :param char: 음절
        :return: 모음 여부
        """
        return char in cls.MO_LIST

    @classmethod
    def is_jaum(cls, char):
        """
        :param char: 음절
        :return: 자음 여부
        """
        return char in cls.JA_LIST

    @classmethod
    def is_cho(cls, char):
        """
        :param char: 음절
        :return: 초성 여부
        """
        return char in cls.CHO_LIST

    @classmethod
    def is_jung(cls, char):
        """
        :param char: 음절
        :return: 종성 여부
        """
        return cls.is_moum(char)

    @classmethod
    def is_jong(cls, char):
        """
        :param char: 음절
        :return: 종성 여부
        """
        return char in cls.JONG_LIST

    @classmethod
    def get_cho(cls, char):
        """
        :param char: 음절
        :return: 초성
        """
        try:
            if cls.is_cho(char):
                return char
            elif cls.is_jung(char) or cls.is_jong(char):
                return ''
            return cls.CHO_LIST[int(cls.__char_offset(char) / int(len(cls.MO_LIST) * len(cls.JONG_LIST)))]
        except:
            return ''

    @classmethod
    def get_jung(cls, char):
        """
        :param char: 음절
        :return: 중성
        """
        if cls.is_jung(char):
            return char
        elif cls.is_cho(char) or cls.is_jong(char):
            return ''
        return cls.MO_LIST[int(cls.__char_offset(char) / int(len(cls.JONG_LIST)) % len(cls.MO_LIST))]

    @classmethod
    def get_jong(cls, char):
        """
        :param char: 음절
        :return: 종성
        """
        if cls.has_jung(char) and cls.is_jong(char):
            return char
        elif cls.is_cho(char) or cls.is_jung(char):
            return ''
        else:
            return cls.JONG_LIST[cls.__char_offset(char) % len(cls.JONG_LIST)]

    @classmethod
    def has_cho(cls, char):
        """
        :param char: 음절
        :return: 초성을 포함하는지 (사용할 일이 있을까?)
        """
        if len(char) > 1:
            char = char[-1]
        return True if len(cls.get_cho(char)) else False

    @classmethod
    def has_jung(cls, char):
        """
        :param char: 음절
        :return: 중성을 포함하는지
        """
        if len(char) > 1:
            char = char[-1]
        return True if len(cls.get_jung(char)) else False

    @classmethod
    def has_jong(cls, char):
        """
        :param char: 음절
        :return: 종성을 포함하는지
        """
        if len(char) > 1:
            char = char[-1]
        return True if len(cls.get_jong(char)) else False

    @classmethod
    def split2cho_jung_jong(cls, char):
        """
        :param char: 음절
        :returns: tuple(초성, 중성, 종성)
        """
        cho = cls.get_cho(char)
        jung = cls.get_jung(char)
        jong = cls.get_jong(char)
        return [cho, jung, jong]

    @classmethod
    def sentence2jaso(cls, sentence: str) -> str:
        li = []
        for char in sentence:
            if cls.is_hangul_char(char):
                cho, jung, jong = cls.split2cho_jung_jong(char)
                if len(cho) > 0:
                    li.append(cho)
                if len(jung) > 0:
                    li.append(jung)
                if len(jong) > 0:
                    li.append(jong)
            else:
                li.append(char)
        return ''.join(li)

    @classmethod
    def join_cho_jung_jong(cls, cho, jung, jong=' '):
        """
        :param jong: 초성
        :param jung: 중성
        :param cho: 종성
        :return: 음절
        """
        if not (cho and jung):
            return cho or jung

        indexes = [tuple.index(*args) for args in zip(cls.CHO_JUNG_JONG_LIST, (cho, jung, jong))]
        offset = (indexes[0] * len(cls.MO_LIST) + indexes[1]) * len(cls.JONG_LIST) + indexes[2]
        return chr(cls.HANGUL_FIRST + offset)

    @classmethod
    def encode_noise(cls, char, max_retry=10):
        jaso_list = cls.split2cho_jung_jong(char)
        x = random.randint(0, 2)
        ch = jaso_list[x]

        if x == 0:
            target_jaso = cls.CHO_NOISE_LIST
        elif x == 1:
            target_jaso = cls.JUNG_NOISE_LIST
        else:
            target_jaso = cls.JONG_NOISE_LIST

        for _ in range(max_retry):
            randidx = random.randint(0, len(target_jaso) - 1)
            jaso_list[x] = target_jaso[randidx]
            if ch != jaso_list[x]:
                break
        return cls.join_cho_jung_jong(*jaso_list)

    @classmethod
    def join_suffix(cls, char, jong_suffix):
        """
        :param jong_suffix: 음절
        :param char: 종성이면서 접미사 (ㄴ, ㄹ, ....)
        :return: 음절
        """
        cho, jung, jong = cls.split2cho_jung_jong(char)
        if cho != '' and jung != '' and jong == '':
            return cls.join_cho_jung_jong(cho, jung, jong_suffix)
        else:
            return ''.join([char, jong_suffix])

    @classmethod
    def split_string(cls, word):
        """
        :param word: 단어
        :return: list(음절)
        """
        li = []
        for char in word:
            if cls.is_full_hangul(char):
                li.extend(cls.split2cho_jung_jong(char))
            else:
                li.append(char)
        return li

    @classmethod
    def join_string(cls, char_list):
        """
        :param char_list: list(음절)
        :return: 단어
        """
        letters = []
        i = len(char_list) - 1
        try:
            while i >= 0:
                if cls.is_jong(char_list[i]):
                    try:
                        letters.insert(0, cls.join_cho_jung_jong(char_list[i - 2], char_list[i - 1], char_list[i]))
                        i -= 3
                    except:
                        try:
                            letters.insert(0, cls.join_cho_jung_jong(char_list[i - 1], char_list[i]))
                            i -= 2
                        except:
                            letters.insert(0, char_list[i])
                            i -= 1
                elif cls.is_jung(char_list[i]):
                    try:
                        letters.insert(0, cls.join_cho_jung_jong(char_list[i - 1], char_list[i]))
                        i -= 2
                    except:
                        letters.insert(0, char_list[i])
                        i -= 1
                else:
                    letters.insert(0, char_list[i])
                    i -= 1
            return u''.join(letters)
        except:
            return None

    # noinspection PyPep8Naming
    @classmethod
    def qwerty_to_hangul(cls, word):
        """
        오타 인식용
        :param word:
        :return:
        """
        chars = []
        for char in word:
            if char in cls.KEYBOARD_ENG_TO_HAN:
                chars.append(cls.KEYBOARD_ENG_TO_HAN[char])
            else:
                chars.append(char)
        hangul = cls.join_string(chars)
        if hangul:
            return hangul
        else:
            return word  # convert failed

    @staticmethod
    def get_except_hangul(chars):
        """한글을 제외한 문자열을 리턴함"""
        li = []
        for char in chars:
            if not HangulUtil.is_hangul_char(char):
                li.append(char)
        return ''.join(li)

    @staticmethod
    def get_except_english(chars):
        """영어를 제외한 문자열을 리턴함"""
        li = []
        for char in chars:
            if not HangulUtil.is_english_char(char):
                li.append(char)
        return ''.join(li)

    @classmethod
    def text2sentences(cls, text: str, sentence_delim='다.', remove_only_one_word=True, has_hangul=True, remove_markdown=True):
        if isinstance(text, str) and len(sentence_delim) > 0:  # split lines with delimiter
            sentences = []
            lines = text.split(sentence_delim)
            for i, s in enumerate(lines):
                s = StringUtil.merge_to_one_line(s)
                if remove_markdown and '|' in s:
                    continue
                if remove_only_one_word and s.count(' ') == 0:  # only one word in a sentence
                    continue
                if has_hangul and not HangulUtil.has_hangul(s):
                    continue
                if remove_markdown and s.startswith('{{'):
                    continue

                if i + 1 != len(lines):
                    s += sentence_delim
                sentences.append(s)
            return sentences
        else:
            return []


HangulUtil.load()

if __name__ == '__main__':
    print(HangulUtil.is_hangul_char('ㅠ'))
    print(HangulUtil.sentence2jaso('ㅠㅠ형대기술로 이런 범죄가 가능한가? 가까운 미래 SF영화 같네...'))
    # print('"%s"' % str(HangulUtil.CHO_LIST))
    # print('"%s"' % str(HangulUtil.JUNG_LIST))
    # print('"%s"' % str(HangulUtil.JONG_LIST))
    # for c in ['한', '사']:
    #     jaso_list = HangulUtil.split2cho_jung_jong(c)
    #     _c = HangulUtil.encode_noise(c)
    #     print(c, '->', jaso_list, '->', _c)
    # text = r'''이 숫염소의 이름은 쾰른의 레전드 선수이자 나중에 명감독으로 꼽힌 헤네스 바이스바일러에서 유래하였다.
    # 숫염소 헤네스는 인근의 쾰른 서커스단으로부터 기증받은 것이다.
    # {{축구 클럽팀 정보2| 클럽 이름 = 1. FSV 마인츠 05| 풀 네임 = 1. Fußball- und Sport-Verein Mainz 05 e.V.| 별칭 = Die Nullfünfer (05년) Karnevalsverein (카니발 클럽)| 설립연도 = 1905년 3월 27일| 홈구장 = 코파스 아레나| 수용인원 = 33,500| 회장 = {{국기그림|독일}} 하랄트 슈트르츠| 스포르팅 매니저 = {{국기그림|독일}} 크리스티안 하이델| 감독 = {{국기그림|스위스}} 마틴 슈미트| 리그 = 1 분데스리가| 시즌 = 2013-14| 순위 = 7위| pattern_la1=_fsvmainz1415h| pattern_b1=_fsvmainz1415h| pattern_ra1=_fsvmainz1415h| pattern_sh1
    # | pattern_so1'''
    #     print(HangulUtil.text2sentences(text))
    # print(len(HangulUtil.HANGUL_LIST))
    # print(len(HangulUtil.to_one_hot_vector('ㄱ')))
    # print(len(HangulUtil.to_one_hot_vector('ㄴ')))
    # print(HangulUtil.text2sentences('사과나무()는 장미목 장미과 배나무아과 사과나무속에 딸린 종이다. 그 열매는 사과(沙果; 砂果)라 하며, 세계적으로 가장 널리 재배되는 과일 품종 가운데 하나이다. 사전적으로 평과(苹果)라고도 한다.'))
    # print(len(HangulUtil.CHAR2ONE_HOT))
    # print(HangulUtil.to_one_hot_vector('ㄲ'))
    # print(HangulUtil.to_one_hot_vector('?'))
    # print(HangulUtil.split_char('길'))
    # print(numpy.concatenate((numpy.zeros(3), numpy.ones(4))))
    # print(HangulUtil.to_cho_jung_jong_vector('ㄱ'))
    # print(HangulUtil.to_cho_jung_jong_vector('각'))
    # print(HangulUtil.to_cho_jung_jong_vector('?'))

    pass

# for ch in HangulUtil.HANGUL_RANGE:
#     print(chr(ch), )
# print(HangulUtil.HANGUL_LIST)
# print(HangulUtil.ENGLISH_LOWER_RANGE)
# print(HangulUtil.ENGLISH_UPPER_RANGE)
# print(HangulUtil.has_english('한f ㅎㅎㅎㅎㅎ'))
# print('ㅌㅌㅌㅌ')
# print('%s, %s, %s' % (HangulUtil.get_cho('ㄹ'), HangulUtil.get_jung('ㄹ'), HangulUtil.get_jong('ㄹ')))
# print('%s, %s, %s' % (HangulUtil.get_cho('ㅏ'), HangulUtil.get_jung('ㅏ'), HangulUtil.get_jong('ㅏ')))
# print('%s, %s, %s' % (HangulUtil.get_cho('라'), HangulUtil.get_jung('라'), HangulUtil.get_jong('라')))
# print('%s, %s, %s' % (HangulUtil.get_cho('란'), HangulUtil.get_jung('란'), HangulUtil.get_jong('란')))
# print(HangulUtil.has_jung('ㄹ'))
# print(HangulUtil.is_full_hangul('하 바'))
# print(HangulUtil.has_hanja('벌크선社'[-1]))
# print(HangulUtil.has_hanja('수출입銀'[-1]))
# print(HangulUtil.is_full_hangul('합병한다.'))
# print(HangulUtil.is_full_hangul('차세대-폐렴구균백신'))
# print(HangulUtil.is_full_hangul_or_english("""[헤럴드경제=박혜림 기자] '오늘의 날씨'만 철썩같이 믿고 어제보다 덜 추운 줄 알고 얇게 입고 나왔다가 낭패보신 경험 있으시죠?""", exclude_chars="""=[]\"'.?!"""))
# print(HangulUtil.is_full_hangul_or_english("""[헤럴드경제=박혜림 기자] '오늘의 날씨'만 철썩같이 믿고 어제보다 덜 추운 줄 알고 얇게 입고 나왔다가 낭패보신 경험 있으시죠?""", exclude_chars="""\"'.?!"""))